/* calltrack - program to get a connect to a track server */
#ifndef lint
static char rcsid[] = "$Id: calltrack.c,v 1.11 1992/10/21 15:05:08 ken Exp $";
#endif

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/file.h>
#include <errno.h>
#include <netdb.h>

#include <fcntl.h>

#include "sig.h"
#include "libc.h"

#ifndef DEBUG
#define TRACKSERVICE	"track"
#else
#define TRACKSERVICE	"tracktest"
#endif

char *progname;

int errno;
SIG_RET_TYPE sendsig();
int des2;

#define LINELEN 256

/* when you get a signal, send the number down stderr */
SIG_RET_TYPE
sendsig(num)
int num;
{
	char c;

	(void) signal(num, sendsig);	/* needed for non-BSD, doesn't hurt */
	c = num;
	write(des2, &c, 1);
}


int
main(argc, argv)
int argc;
char **argv;
{
	char *host, buf[BUFSIZ], args[LINELEN] ;
	register char *p;
	int des1, childid, count;
	char *hname = NULL;
	fd_set readfrom, fds;
	struct servent *sp;
	int nfds, one = 1;
	int c, errs = 0;
	SIGHANDLER int_proc, quit_proc, term_proc;
	void tackon();


	args[0] =  '\0';
	progname = p = argv[0];
	while (*p) 
		if (*p++ == '/') 
			progname = p;
	/*
	 * the allowed arguments are -s -q and -l <librarian> the first non
	 * flagged word is taken as the subscription list name. anything
	 * following it are ignored.
	 */
	while ((c = getopt(argc, argv, "sql:")) != EOF) {
		switch (c) {
		case 'l':
			hname = optarg;
			break;
		case 's':
			tackon(args, sizeof(args), "-s");
			break;
		case 'q':
			tackon(args, sizeof(args), "-q");
			break;
		case '?':
			errs++;
			break;
		}
	}
	if (errs > 0 || optind != argc - 1) {
		fprintf(stderr, "Usage: %s [-s] [-q] [-l librarian] slist\n",
			progname);
		exit(EINVAL);
	}
	tackon(args, sizeof(args), argv[optind]);
	sp = getservbyname(TRACKSERVICE, "tcp");
	if (sp == 0) {
		fprintf(stderr, "%s: %s/tcp: unknown service\n", progname,
			TRACKSERVICE);
		exit(ENOENT);
	}
	/*
	 * don't bother sending any reasonable names for users, trackd ignores
	 * them
	 */
	des1 = rcmd(&hname, sp->s_port, "foo", "foo", args, &des2);
	if (des1 < 0) {
		exit(errno ? errno : EIO);
	}

	if (des2 < 0) {
		fprintf(stderr, "%s: can't establish stderr\n", progname);
		exit(errno ? errno : EIO);
	}

	c = getuid();
	if (setuid(c) < 0 || getuid(c) != c) {	/* SHOULDNTHAPPEN */
		/* SCO Unix or some suchlike system with "security"? */
		perror("calltrack: setuid(getuid()) failed");
		exit(errno ? errno : EACCES);
	}
	if ((quit_proc = signal(SIGQUIT, SIG_IGN)) != SIG_IGN)
		(void) signal(SIGQUIT, sendsig);
	if ((int_proc = signal(SIGINT, SIG_IGN)) != SIG_IGN)
		(void) signal(SIGINT, sendsig);
	if ((term_proc = signal(SIGTERM, SIG_IGN)) != SIG_IGN)
		(void) signal(SIGTERM, sendsig);

#ifdef O_NONBLOCK
	fcntl(des1, F_SETFL, O_NONBLOCK);
#else
	ioctl(des1, FIONBIO, &one);
#endif
	childid = fork();
	if (childid < 0) {
		perror("calltrack: fork");
		exit(errno ? errno : EAGAIN);
	}
	if (childid == 0) {
		/* this process will write to the socket 0 -> des1 */
		char *bufptr; 
		int wcount;

		close(des2);
		des2 = -1;
readloop:
		errno = 0;
		count = read(0, buf, sizeof buf);
		if (count <= 0) {
			if (count < 0)
				perror("calltrack: read");
			goto done;
		}
		bufptr = buf;
writeloop:
		FD_ZERO(&fds);
		FD_SET(des1, &fds);
		if (select(32, 0, &fds, 0, 0) < 0) {
			if (errno != EINTR) {
				perror("calltrack: select");
				exit(errno ? errno : EIO);
			}
			goto writeloop;
		}
		if (!FD_ISSET(des1, &fds))
			goto writeloop;
		wcount = write(des1, bufptr, count);
		if (wcount < 0) {
			if (errno == EWOULDBLOCK) {
				goto writeloop;
			}
			perror("calltrack: write");
			goto done;
		}
		bufptr += wcount;
		count -= wcount;
		if (count == 0) {
			goto readloop;
		}
		goto writeloop;
done:
		shutdown(des1, 1);
		exit(0);
	}
	/*
	 * restore old values - the *_proc could be SIG_IGN if this was
	 * running in the background from a Bourne shell.
	 */
	(void) signal(SIGQUIT, quit_proc);
	(void) signal(SIGINT, int_proc);
	(void) signal(SIGTERM, term_proc);
	FD_ZERO(&readfrom);
	FD_SET(des1, &readfrom);
	FD_SET(des2, &readfrom);
	nfds = 2;
	do {
		/*
		 * this process will read from the socket: des2 -> 2, des1 -> 1
		 */
		fds = readfrom;
		if (select(32, &fds, 0, 0, 0) < 0) {
			if (errno != EINTR) {
				perror("select");
				exit(errno ? errno : EIO);
			}
			continue;
		}
		/* anything on standard error? */
		if (FD_ISSET(des2, &fds)) {
			errno = 0;
			count = read(des2, buf, sizeof buf);
			if (count <= 0) {
				if (errno != EWOULDBLOCK) {
					if (count < 0)
					    perror("calltrack: read stderr");
					FD_CLR(des2, &readfrom);
					nfds--;
				}
			} else {
				pwrite(2, buf, count);
			}
		}
		/* anything on standard input? */
		if (FD_ISSET(des1, &fds)) {
			errno = 0;
			count = read(des1, buf, sizeof buf);
			if (count <= 0) {
				if (errno != EWOULDBLOCK) {
					if (count < 0)
					    perror("calltrack: read stdin");
					FD_CLR(des1, &readfrom);
					nfds--;
				}
			} else
				pwrite(1, buf, count);
		}
	} while (nfds);
	/* kill the other process */
	kill(childid, SIGKILL);
	return 0;
}


void
tackon(to, size, from)
register char *to, *from;
int size;
{
	register char *p;

	for (p = to; *p; p++)
		;
	/*
	 * p - to accounts for what is already there (less the null byte).
	 * The extra 2 accounts for the null byte and the blank.
	 */
	size -= (p - to) - 2 - strlen(from);
	if (size >= 0) {
		*p++ = ' ';
		(void) strcpy(p, from);
	} else {
		(void) fprintf(stderr, "%s: buffer overflow - %s\n", progname, to);
		exit(EFBIG);
	}
	return;
}


pwrite(fd, p, n)
char *p;
{
	int cc;

	while (n > 0) {
		cc = write(fd, p, n);
		if (cc <= 0) {
			perror("calltrack: write");
			exit(errno ? errno : EIO);
		}
		n -= cc;
		p += cc;
	}
}


