/* communcations routines */
#ifndef lint
static char rcsid[] = "$Id: comm.c,v 1.16 1994/05/30 16:52:14 root Exp $";
#endif

#include "track.h"
#include <errno.h>
#include "sig.h"
#include <sys/wait.h>

int outpipe = -1;
int inpipe  = -1;

int clobber = NO_CLOBBER;

static SIG_RET_TYPE
tcphung()
{
	sprintf(errmsg, "tcp connection timed out after %d seconds", TCP_TIME);
	do_gripe();
	clearlock();
	/* kill the tcp processes */
	kill(child_id, SIGTERM);
	exit(1);
}

/*
 * Return exit status of calltrack process, or -1 if it hasn't died yet.
 */
reapchild()
{
        int status;

	return waitpid((pid_t)child_id, &status, WNOHANG) > 0 ? status : -1;
}

startget(clobberflag, listflag)
int clobberflag, listflag;
{
	int p1[2], p2[2];
	char rcmd[LINELEN];
	char qstring[WORDLEN];
	char lstring[WORDLEN];
	char userstring[WORDLEN];
	char *args[20];			/* XXX */
	char cmd[LINELEN];

	if ((outpipe >= 0) || (inpipe >= 0)) {
		strcpy(errmsg, "startget called while already connected");
		do_panic();
	}

	strcpy(qstring, "");
	if (quietflag) {
		strcpy(qstring, "-q");
	}
	strcpy(lstring, "");
	if (listflag == IS_LIST) {
		strcpy(lstring, "-s");
	}
	strcpy(userstring, "");
	if (strlen(user) != 0) {
		sprintf(userstring, "-l %s", user);
	}

	switch (via) {
	case VIA_TCP:
#ifndef DEBUG
		sprintf(rcmd, "%s/bin/calltrack -l %s %s %s %s",
			rootdir, librarian, qstring, lstring, subname);
#else
		sprintf(rcmd, "%s/calltrack -l %s %s %s %s",
			SRCDIR, qstring, lstring, subname);
#endif
		break;
	default:
		sprintf(rcmd, 
			"/usr/ucb/rsh %s %s %s/bin/track -d%s -g%s %s %s %s",
		     	librarian, userstring, rrootdir, rrootdir, 
			myname, qstring, lstring, subname);
		break;
	}
	doargs(args, sizeof args / sizeof args[0], rcmd);
	strcpy(cmd, args[0]);

	/*
	 * set an alarm so that if something hangs we'll eventually clean up.
	 * this is most useful when the librarian is running, but has a jammed
	 * process table.
	 */
	if ((via == VIA_TCP) || (via == VIA_RSH)) {
		signal(SIGALRM, tcphung);
		alarm(TCP_TIME);
	}

	if (verboseflag)
		printf("execing -- %s\n", rcmd);
	sprintf(errmsg, "can't open pipe for %s", rcmd);

	if (pipe(p1) < 0)
		do_panic();
	if (pipe(p2) < 0)
		do_panic();
	if ((child_id = fork()) == 0) {

		dup2(p1[0], 0);

		dup2(p2[1], 1);
		dup2(1, 2);		/* seems wrong... */

		close(p1[0]);
		close(p1[1]);

		close(p2[0]);
		close(p2[1]);

		execv(cmd, args);
		_exit(1);
	}
	if (child_id == -1)
		do_panic();
	outpipe = p1[1];
	inpipe  = p2[0];
	close(p1[0]);
	close(p2[1]);

	strcpy(errmsg, "");
	clobber = clobberflag;
}


stopget()
{
	int tmpstat;
	if (((tmpstat = mypclose()) == -1) || (((tmpstat >> 8) & 0377) != 0)) {
		sprintf(errmsg, "got funny status from getpipe %d", tmpstat);
		do_gripe();
		alarm(0);
		return(-1);
	}
	alarm(0);
	return(0);
}


mypclose()
{
	register r;
	SIGHANDLER hstat, istat, qstat;
	int status;

	close(outpipe);
	outpipe = -1;
	istat = signal(SIGINT, SIG_IGN);
	qstat = signal(SIGQUIT, SIG_IGN);
	hstat = signal(SIGHUP, SIG_IGN);
	while ((r = wait(&status)) != child_id && r != -1)
		;
	if (r == -1) {
		if (errno == ECHILD) {
			/* one of the other pcloses reaped it */
			status = 0;
			errno = 0;
		} else {
			status = -1;
		}
	}
	signal(SIGINT, istat);
	signal(SIGQUIT, qstat);
	signal(SIGHUP, hstat);
	close(inpipe);
	inpipe = -1;
	return(status);
}


doargs(arr, arrsz, ptr)
char *arr[], *ptr;
{
	int i;
	char buf[LINELEN];
	int len;

	skipspace(&ptr);
	for (i = 0; *ptr != '\0'; i++) {
		len = mycpy(buf, sizeof buf, ptr) - buf;
		if (i >= arrsz) {
			sprintf(errmsg, "too many args (%d) to command, max. %d\n",
				i, arrsz);
			do_panic();
		}
		arr[i] = strcpy((char *) emalloc(len + 1), buf);
		skipwordspace(&ptr);
	}
	arr[i] = (char * )0;
}
