/* routines that handle file transfer for librarian and subscriber */
#ifndef lint
static char rcsid[] = "$Id: copyfile.c,v 1.25 92/06/17 18:38:24 ken Exp $";
#endif

#include "track.h"
#include "pgcache.h"
#include <errno.h>

#define min(x,y)	( (x) < (y) ? (x) : (y) )

extern int clobber;	/* from comm.c */

/*
 * When the subscriber has sent us a file request, and is waiting for a
 * reply in printhead() format, we must send it an error message of the
 * form "sep_char error" if we cannot send it the head for some reason.
 */
copyerror(s)
char *s;
{
	putchar(sep_char);
	puts(s);
	fflush(stdout);
}


int
printhead(name, outname, linkflag)
char *name, *outname;
int linkflag;
{
	struct stat sbuf;
	struct passwd *pw = NULL;
	struct group *gr = NULL;
	char linkbuf[LINELEN];
	int linklen;
	int type;	/* file type	*/

	if ((type = gettype_stat(name, &sbuf)) != -1) {
		if ((pw = getpwuid_cache(sbuf.st_uid)) == NULL) {
			sprintf(errmsg, "printhead: getpwuid(%d) failed",
				sbuf.st_uid);
			copyerror(errmsg);
			do_gripe();
			return -1;
		}
		if ((gr = getgrgid_cache(sbuf.st_gid)) == NULL) {
			sprintf(errmsg, "printhead: getgrgid(%d) failed",
				sbuf.st_gid);
			copyerror(errmsg);
			do_gripe();
			return -1;
		}
	}
	switch (type) {
	case 'f':
		if (linkflag) {
			sbuf.st_size = (long) 0;
		}
		/* FALLTHROUGH */
	case 'd':
		printf("%s%c%o%c%s%c%s%c%ld%c%ld%c%ld%c%d%c%ld%c%ld%c%ld%c\n",
		       outname, sep_char, sbuf.st_mode & MODE_BITS,
		       sep_char, pw->pw_name,
		       sep_char, gr->gr_name, sep_char, (long) sbuf.st_atime,
		       sep_char, (long) sbuf.st_mtime, sep_char,
		       (long) sbuf.st_ctime, sep_char, sbuf.st_nlink,
		       sep_char, (long) sbuf.st_dev, sep_char,
		       (long) sbuf.st_ino, sep_char, (long) sbuf.st_size,
		       sep_char);
		break;
	case 'l':
		if ((linklen = readlink(name, linkbuf, LINELEN)) == -1) {
			sprintf(errmsg, "readlink(%s) failed ", name);
			copyerror(errmsg);
			do_gripe();
			return -1;
		}
		/* null terminate the link name */
		linkbuf[linklen] = '\0';
		printf("%s%c%o%c%s%c%s%c", outname, sep_char,
		       sbuf.st_mode & MODE_BITS, sep_char,
		       pw->pw_name, sep_char, gr->gr_name, sep_char);

		printf("%ld%c%ld%c%ld%c%d%c%ld%c%ld%c%s%c\n",
		       (long) sbuf.st_atime, sep_char, (long) sbuf.st_mtime,
		       sep_char, (long) sbuf.st_ctime, sep_char,
		       sbuf.st_nlink, sep_char, (long) sbuf.st_dev, sep_char,
		       (long) sbuf.st_ino, sep_char, linkbuf, sep_char);
		break;
	case -1:
		sprintf(errmsg, "printhead can't find file named %s", name);
		copyerror(errmsg);
		do_gripe();
		return -1;
	default:	/* CANTHAPPEN */
		sprintf(errmsg, "printhead: gettype_stat(%s) = %d", name,
			type);
		copyerror(errmsg);
		do_gripe();
		return -1;
	}
	fflush(stdout);
	return 0;
}


typedef struct link_str {
	dev_t l_dev;
	ino_t l_ino;
	char *l_name;
	struct link_str *l_next;
} LinkStruct;

char *
linkto(nameptr, dev, ino)
char *nameptr;
dev_t dev;
ino_t ino;
{
	static LinkStruct *the_links;
	LinkStruct *lp;

	for(lp = the_links; lp != NULL; lp = lp->l_next)
		if (lp->l_ino == ino && lp->l_dev == dev)
			return(lp->l_name);
	lp = (LinkStruct *) emalloc(sizeof(LinkStruct));
	lp->l_dev = dev;
	lp->l_ino = ino;
	lp->l_name = strsave(nameptr);
	lp->l_next = the_links;
	the_links = lp;
	return((char * )0);
}


writefile(outname, buf)
char *buf, *outname;
{
	char *ptr;
	int infile;
	struct stat st;

	/* find and isolate the first word on the line */
	ptr = buf;
	skipspace(&ptr);
	skipword(&ptr);
	*ptr = '\0';
	ptr = buf;

	skipspace(&ptr);
	if (strlen(ptr) == 0) {
		sprintf(errmsg, "writefile got weird line :%s:", buf);
		do_panic();
		/* NOTREACHED */
	}

	switch (gettype_stat(ptr, &st)) {
	case -1:
		sprintf(errmsg, "can't find file %s", ptr);
		copyerror(errmsg);
		do_gripe();
		return;
	case 'f':
		if ((infile = open(ptr, 0)) == -1) {
			sprintf(errmsg, "can't open file named %s", ptr);
			copyerror(errmsg);
			do_gripe();
		} else {
			if (st.st_nlink == 1 ||
			    linkto(outname, st.st_dev, st.st_ino) == 0) {
				if (printhead(ptr, outname, 0) == 0) {
					passbytes(infile, outpipe, st.st_size);
				}
			} else {
				printhead(ptr, outname, 1);
			}
			if (close(infile) == -1) {
				sprintf(errmsg, "couldn't close(%s)", ptr);
				do_gripe();
			}
		}
		break;
	default:
		printhead(ptr, outname, 0);
	}
}


getfield(ptr, buf)
char **ptr, *buf;
{
	register char *to = buf;
	register char *cp = *ptr;
	register char c;

	while ((c = *cp++) != '\0') {
		if (c == sep_char) {
			*to = '\0';
			*ptr = cp;
			return;
		}
		*to++ = c;
	}
	sprintf(errmsg, "getfield ran off end of line working on :%s:", buf);
	do_panic();
}


long
gettime(buf)
char *buf;
{
	char *ptr;
	char dummy[LINELEN];

	ptr = buf;

	getfield(&ptr, dummy);	/* skip name  */
	getfield(&ptr, dummy);	/*  "   modes */
	getfield(&ptr, dummy);	/*  "   owner */
	getfield(&ptr, dummy);	/*  "   group */
	getfield(&ptr, dummy);	/*  "   atime */
	getfield(&ptr, dummy);	/* get mtime */
	return(atol(dummy));
}


long
getsize(buf)
char *buf;
{
	char *ptr;
	char dummy[LINELEN];

	ptr = buf;

	getfield(&ptr, dummy);	/* skip name  */
	getfield(&ptr, dummy);	/*  "   modes */
	getfield(&ptr, dummy);	/*  "   owner */
	getfield(&ptr, dummy);	/*  "   group */
	getfield(&ptr, dummy);	/*  "   atime */
	getfield(&ptr, dummy);	/*  "   mtime */
	getfield(&ptr, dummy);	/*  "   ctime */
	getfield(&ptr, dummy);	/*  "   nlinks*/
	getfield(&ptr, dummy);	/*  "   devic */
	getfield(&ptr, dummy);	/*  "   inode */

	getfield(&ptr, dummy);	/* get  size  */
	if ( *ptr != '\0') {
		sprintf(errmsg, "getfield didn't find end of line in :%s:",
		    buf);
		do_panic();
	}
	return(atol(dummy));
}


flush_body(buf)
char *buf;
{
	char namebuf[LINELEN], modebuf[LINELEN];
	char *ptr;
	int mode;

	/* find and isolate the first word on the line */
	ptr = buf;
	getfield(&ptr, namebuf);
	getfield(&ptr, modebuf);
	if (sscanf(modebuf, "%o", &mode)  != 1) {
		sprintf(errmsg, "couldn't scanf mode :%s:", modebuf);
		do_panic();
	}
	switch (mode & S_IFMT) {
	case S_IFDIR:
	case S_IFLNK:
		break;
	case S_IFREG:
		skipbytes(inpipe, getsize(buf));
		break;
	}
}


getmyuid(name)
char *name;
{
	struct passwd *pw;

	if ((pw = getpwnam_cache(name)) != NULL)
		return(pw->pw_uid);
	sprintf(errmsg, " can't find uid for user %s using uid = %d", name,
	    geteuid());
	return(geteuid());
}


getmygid(name)
char *name;
{
	struct group *gr;

	if ((gr = getgrnam_cache(name)) != (struct group *) NULL)
		return(gr->gr_gid);
	sprintf(errmsg, " can't find gid for group %s using gid = %d", name,
	     getegid());
	return(getegid());
}


/*
 * my version of fgets, but doesn't use buffering so it won't collide with
 * other reads.  !! Should implement a decent buffering scheme for inpipe so
 * we don't have to do these 1 byte reads!  Returns the number of bytes read.
 */
int
get_head(ptr, sz)
register char *ptr;
int sz;
{
	char bufc;
	register int i = 0;

	if (sz <= 0) {
		strcpy(errmsg, "get_head called with sz == 0");
		do_panic();
	}
	sz--;
	for (;;) {
		switch (read(inpipe, &bufc, 1)) {
		case -1:
			if (errno == EINTR)
				continue;
			strcpy(errmsg, "read error");
			do_panic();
		case 0:
			bufc = '\n';
		}
		if (bufc == '\n') {
			*ptr = '\0';
			return i;
		}
		if (i >= sz) {
			strcpy(errmsg, "get_head buffer overflow");
			do_panic();
		}
		i++;
		*ptr++ = bufc;
	}
}


int
get_body(buf, toname)
char *buf;
char *toname;
{
	char sbuf[LINELEN];
	char tmpbuf[LINELEN];
#define TMPFILENAME "#track.tmp.XXXXXX"
	char tempfile[LINELEN + sizeof TMPFILENAME];
	char *name1;
	char *ptr;
	unsigned long timevec[2];
	int outfile, status;
	int usave;
	int type;	/* file type	*/
	int errs = 0;
	int mode, uid, gid;
	dev_t dev;
	ino_t ino;
	int nlink;
	long size;
	struct stat stbuf;
	Entry *ep;

	usave = umask(000);

	/* parse the header */
	ptr = buf;
	getfield(&ptr, tmpbuf);

	getfield(&ptr, tmpbuf);
	if (sscanf(tmpbuf, "%o", &mode) != 1) {
		sprintf(errmsg, "writefile couldn't scanf mode :%s:", tmpbuf);
		do_panic();
		/* NOTREACHED */
	}

	getfield(&ptr, tmpbuf);
	uid = getmyuid(tmpbuf);

	getfield(&ptr, tmpbuf);
	gid = getmygid(tmpbuf);

	getfield(&ptr, tmpbuf);
	timevec[0] = atol(tmpbuf);

	getfield(&ptr, tmpbuf);
	timevec[1] = atol(tmpbuf);

	getfield(&ptr, tmpbuf);
	/* ctime not used yet */

	getfield(&ptr, tmpbuf);
	nlink = atoi(tmpbuf);

	getfield(&ptr, tmpbuf);
	dev = atoi(tmpbuf);

	getfield(&ptr, tmpbuf);
	ino = atoi(tmpbuf);

	getfield(&ptr, sbuf);

	if (*ptr != '\0') {
		sprintf(errmsg, "didn't find EOL in :%s:", buf);
		do_panic();
		/* NOTREACHED */
	}

	if (checkroot(toname) == 0) {
		if (dirflag) {
			makeroot(toname, uid, gid, mode);
		} else {
			sprintf(errmsg, "can't find parent directory for %s",
			     	toname);
			do_gripe();
			errs++;
			if ((mode & S_IFMT) == S_IFREG)
				skipbytes(inpipe, atol(sbuf));
			goto done;
		}
	}
	switch (mode & S_IFMT) {
	case S_IFLNK:
		switch (type = gettype_stat(toname, &stbuf)) {
		case 'd':
		case 'f':
			if (forceflag) {
				if (removeit(toname, type) == -1) {
					sprintf(errmsg,
						"can't remove file at %s",
						toname);
					do_panic();
					/* NOTREACHED */
				}
			} else {
				sprintf(errmsg, "%s not a link  -- skipping",
				     	toname);
				errno = 0;
				do_gripe();
				errs++;
				goto done;
			}
			break;
		case 'l' :
			if (removeit(toname, type) == -1) {
				sprintf(errmsg, "can't remove link at %s",
					toname);
				do_gripe();
				errs++;
				goto done;
			}
		case -1 :
			break;
		default :	/* CANTHAPPEN */
			sprintf(errmsg, "got funny code %d for %s", type,
				toname);
			do_gripe();
			errs++;
			goto done;
		}
		/* 
		 * set the umask, chmod doesn't work right on symbolic
		 * links
		 */
		umask(~(mode & 0777));
		ep = &entries[cur_ent];
		(void) remapname(ep, sbuf, sizeof sbuf, strlen(sbuf));
		if (symlink(sbuf, toname)) {
			sprintf(errmsg, "can't create symbolic link from %s to %s",
			     	toname, sbuf);
			do_gripe();
			errs++;
			goto done;
		}
		umask(000);
		break;

	case S_IFDIR:
		switch (type = gettype_stat(toname, &stbuf)) {
		case 'd' :
		case -1 :
			break;
		case 'f':
		case 'l':
			if (forceflag) {
				if (removeit(toname, type) == -1) {
					sprintf(errmsg, "can't remove %s",
						toname);
					do_panic();
					/* NOTREACHED */
				}
			} else {
				sprintf(errmsg, "%s not a directory  -- skipping",
				     	toname);
				errno = 0;
				do_gripe();
				errs++;
				goto done;
			}
			break;
		default :		/* CANTHAPPEN */
			sprintf(errmsg, "got funny code %d for %s", type,
				toname);
			do_gripe();
			errs++;
			goto done;
		}
		if ( type == -1 && (mkdir(toname, mode) == -1) ) {
			sprintf(errmsg, "cant create directory %s", toname);
			do_gripe();
			errs++;
			goto done;
		}
		if (uid != stbuf.st_uid || gid != stbuf.st_gid) {
			if (chown(toname, uid, gid) == -1) {
				sprintf(errmsg,
					"can't chown directory %s %d %d\n",
				     	toname, uid, gid);
				do_gripe();
				goto done;
			}
		}
		if ( chmod(toname, mode) == -1) {
			sprintf(errmsg,
				"can't chmod directory %s to mode 0%03o\n",
			     	toname, mode);
			do_gripe();
			goto done;
		}
		break;

	case S_IFREG:
		size = atol(sbuf);
		switch (type = gettype_stat(toname, &stbuf)) {
		case 'f':
		case -1 :
			break;
		case 'd':
		case 'l':
			if (forceflag) {
				if (removeit(toname, type) == -1) {
					sprintf(errmsg, "can't remove %s",
						toname);
					do_panic();
					/* NOTREACHED */
				}
			} else {
				sprintf(errmsg,
					"%s not a regular file -- skipping",
				     	toname);
				errno = 0;
				do_gripe();
				errs++;
				skipbytes(inpipe, size);
				goto done;
			}
			break;
		default :		/* CANTHAPPEN */
			sprintf(errmsg, "got funny code %d for %s", type,
				toname);
			do_gripe();
			errs++;
			skipbytes(inpipe, size);
			goto done;
		}

		/* 
		 * !! We must get all hard links to a file in the same
		 * track session for this algorithm to work.  We panic
		 * because we have entered toname into the_links in
		 * linkto, and cannot afford to leave it like that.
		 * TODO:  split linkto() into ishardlink() and
		 * savehardlink().
		 */
		if ((nlink > 1) && (name1 = linkto(toname, dev, ino))) {
			if (type != -1 && unlink(toname) < 0) {
				sprintf(errmsg, "can't unlink %s", toname);
				do_panic();
				/* NOTREACHED */
			}
			if (link(name1, toname) == -1) {
				sprintf(errmsg, "can't link %s and %s",
					toname, name1);
				do_panic();
				/* NOTREACHED */
			} else {
				if (verboseflag) {
					printf("linked %s to %s\n", name1,
					       toname);
				}
			}
			skipbytes(inpipe, size);
			status = setattr(toname, -1, uid, gid, mode, timevec);
			if (status < 0) {
				strcpy(errmsg, "setattr failed");
				do_panic();
				/* NOTREACHED */
			}
			break;
		}
		/* 
		 * don't clobber newer versions unless told to do so
		 */
		if (!clobber && type != -1 && stbuf.st_mtime > timevec[1]) {
			sprintf(errmsg, "%s is newer", toname);
			errno = 0;
			do_gripe();
			errs++;
			skipbytes(inpipe, size);
			goto done;
		}
		/*
		 * Copy to a temp file, rename when done.
		 */
		strcpy(tempfile, toname);
		ptr = strrchr(tempfile, '/');
		strcpy(ptr ? ptr + 1 : tempfile, TMPFILENAME);

		if ((outfile = opentmp(tempfile)) < 0) {
			sprintf(errmsg, "can't create temp file %s", tempfile);
			do_gripe();
			errs++;
			skipbytes(inpipe, size);
			goto done;
		}
		fchmod(outfile, 0600);	/* conservative */
		passbytes(inpipe, outfile, size);
		status = setattr(tempfile, outfile, uid, gid, mode, timevec);
		if (status < 0) {
			(void) unlink(tempfile);	/* forget it */
			if (nlink > 1) {
				do_panic();
				/* NOTREACHED */
			}
			errs++;
			goto done;
		} else if (rename(tempfile, toname) < 0) {
			(void) unlink(tempfile);
			sprintf(errmsg, "can't rename %s to %s", tempfile,
				toname);
			if (nlink > 1)
				do_panic();
			else
				do_gripe();
			errs++;
			goto done;
		}
		break;
	default:
		sprintf(errmsg, "get_body -- bad file mode %o for %s", mode,
		     	toname);
		do_panic();
	}
done:
	umask(usave);
	return errs;
}


/*
 * Set the given file attributes.  Under some Unixes, close() resets the
 * file's mtime, so we must close the file before setting the times.
 */
setattr(file, fd, uid, gid, mode, times)
char *file;
int fd;			/* open file descriptor for file, or -1 */
int uid, gid, mode;
unsigned long *times;
{
	struct stat st;
	int status = 0;
	time_t t[2];

	if ((fd < 0 ? lstat(file, &st) : fstat(fd, &st)) < 0) {
		sprintf(errmsg, "Can't stat %s", file);
		do_gripe();
		if (fd >= 0)
			close(fd);
		return -1;
	}
	if (st.st_uid != uid || st.st_gid != gid) {
		if ((fd < 0 ? chown(file, uid, gid) : fchown(fd, uid, gid)) <
		    0) {
			sprintf(errmsg, "can't chown file %s to %d/%d", file,
			     uid, gid);
			do_gripe();
			if (fd >= 0)
				close(fd), fd = -1;
			status = -1;
		}
	}
	if ((fd < 0 ? chmod(file, mode) : fchmod(fd, mode)) < 0) {
		sprintf(errmsg, "can't chmod file %s to %x", file, mode);
		do_gripe();
		if (fd >= 0)
			close(fd), fd = -1;
		status = -1;
	}
	if (fd >= 0)
		close(fd);
	t[0] = times[0];
	t[1] = times[1];
	if (utime(file, t) < 0) {
		sprintf(errmsg, "can't set times on %s", file);
		do_gripe();
		status = -1;
	}
	return status;
}


passbytes(from, to, togo)
int from, to;
long togo;
{
	int rcnt;
	long toget;
	char bbuf[BLOCKSIZE];

	/* stdio would probably be more efficient */
	while (togo) {
		toget = min(BLOCKSIZE, togo);
		if ((rcnt = read(from, bbuf, (int) toget)) <= 0) {
			sprintf(errmsg, "passbytes -- read wanted %d, got %d",
			 	toget, rcnt);
			do_panic();
			/* NOTREACHED */
		}
		if (pwrite(to, bbuf, rcnt) != rcnt) {
			sprintf(errmsg, "passybytes -- write error expected %d",
				rcnt);
			do_panic();
			/* NOTREACHED */
		}
		togo -= rcnt;
		if ( togo < 0 ) {
			sprintf(errmsg, "passbytes -- togo went negative");
			do_panic();
			/* NOTREACHED */
		}
	}
}


skipbytes(from, togo)
int from;
long togo;
{
	int rcnt;
	long toget;
	char bbuf[BLOCKSIZE];
	
	while (togo) {
		toget = min(BLOCKSIZE, togo);
		if ((rcnt = read(from, bbuf, (int) toget)) <= 0) {
			sprintf(errmsg, "skipbytes -- read wanted %d, got %d",
				toget, rcnt);
			do_panic();
		}
		togo -= rcnt;
		if ( togo < 0 ) {
			sprintf(errmsg, "skipbytes -- togo went negative");
			do_panic();
		}
	}
}
