/* exception list handling. */
#ifndef lint
static char rcsid[] = "$Id: except.c,v 1.10 91/04/23 12:09:55 root Exp $";
#endif

#include <stdio.h>
#include <ctype.h>
#include "track.h"

/*
 * routine to implement exception lists. returns 1 if the filename passed in
 * theline is matches the string passed in rootstring and does NOT match any
 * of the exceptions found in the entry pointed at by entnum returns 0
 * otherwise
 */
int
goodname(theline, rootstring, entnum)
char *rootstring;
char *theline;
int entnum;
{
	register char *ptr;
	register char *ptr2;

	/* find the first '/' */
	for (ptr2 = theline; *ptr2 != '\0'; ptr2++)
		if (*ptr2 == '/')
			break;
	if (*ptr2 != '/') {
		error("can't find / in %s", theline);
		/*NOTREACHED*/
	}

	/*
	 * scan until white space, that's the filename.  simultaneously
	 * match with rootstring.
	 */
	ptr = rootstring;
	while (*ptr2 == *ptr && *ptr2 != '\0') {
		ptr++;
		ptr2++;
	}
	if (*ptr != '\0')	/* not end of rootstring, so we failed */
		return(0);
	if (*ptr2 == '\0' || isspace(*ptr2))  /* entire match, success */
		return(1);
	if (*ptr2 != '/')	/* failed */
		return(0);

	/*
	 * the head matches, so now we have to check the tail 
	 * in ptr2 (SKIPPING THE / SEPARATING THE HEAD AND TAIL)
	 */
	return ngoodname(ptr2, entnum);
}


/* 
 * headmatch returns 0 if s1 == root, or root followed by '/' is a
 * prefix of s1, a negative number if s1 is lexicographically less than
 * root, and a positive number if s1 is lexicographically greater than
 * root
 */
int
headmatch(s1, root, rootlen)
char *s1, *root;
int rootlen;
{
	char nextchar;
	int i;

	if ((i = strncmp(s1, root, rootlen)) == 0) {
		nextchar = s1[rootlen];
		if (nextchar == '\0' || nextchar == '/' || isspace(nextchar))
			return(0);
		else
			return nextchar;
	}
	return(i);
}


int
ngoodname(fname, entnum)
char *fname;
int entnum;
{
	char savechar;
	register char *ptr;
	register char *ptr2;
	register char **excp;
	register regexp **rep;

	ptr2 = fname;
	if (*ptr2 != '/') /* must be '\0' or a space.  entire match, success */
		return(1);

	excp = &entries[entnum].exceptions[0];
	if (*excp == 0)				/* no exceptions, success */
		return(1);

	ptr = ptr2++;

	while (*ptr != '\0' && !isspace(*ptr))
		ptr++;
	savechar = *ptr;
	*ptr = '\0';

	/*
	 * compare the tail with each exception if there is a match, the
	 * return(0) else return(1)
	 */
	rep = &entries[entnum].re[0];
	for (; *excp != 0; excp++, rep++) {
		if (*rep == 0) {
			if (*ptr2 == **excp &&
			    headmatch(ptr2, *excp, strlen(*excp)) == 0) {
				*ptr = savechar;
				return(0);
			}
		} else {
			if (regexec(*rep, ptr2)) {
				*ptr = savechar;
				return(0) ;
			}
		}
	}
	*ptr = savechar;
	return(1);
}


