/* code to determine file type, make directory hierarchies, resolve symlinks */
#ifndef lint
static char rcsid[] = "$Id: files.c,v 1.13 1993/04/14 17:43:11 lbin Exp $";
#endif

#include "track.h"
#include <errno.h>

gettype_stat(name, st)
char *name;
struct stat *st;
{
	struct stat sbuf;

	if (st == NULL)
		st = &sbuf;

        /* Initialize st_mode since lstat does not set values of
           st if lstat call fails -edwin */
        st->st_mode = 0;

	if (lstat(name, st) == 0) {
		switch (st->st_mode & S_IFMT) {
		case S_IFDIR:
			return('d');
		case S_IFREG:
			return('f');
		case S_IFLNK:
			return('l');
		default:
			sprintf(errmsg, "gettype -- bad file mode for %s",
				name);
			do_panic();
		}
		/* NOTREACHED */
	}
	if (errno == ENOENT)
		return(-1);
	sprintf(errmsg, "gettype_stat -- lstat(%s) failed", name);
	do_panic();
	/* NOTREACHED */
}


char *
resolve(ptr)
char *ptr;
{
	static char retval[LINELEN];
	int linklen;

	if ((linklen = readlink(ptr, retval, LINELEN)) != -1 &&
	    linklen != LINELEN) {
		retval[linklen] = '\0';
	} else {
		sprintf(errmsg, "warning can't follow link named %s", ptr);
		do_panic();
		/* NOTREACHED */
	}
	return(retval);
}


checkroot(ptr)
char *ptr;
{
	char checkbuf[LINELEN], *tmp;

	/* assume / exists */
	if (strcmp(ptr, "/") == 0)
		return(1);
	strcpy(checkbuf, ptr);
	if ((tmp = strrchr(checkbuf, '/')) ==  0) {
		sprintf(errmsg, "checkroot can't find / in %s", checkbuf);
		do_gripe();
		return(0);
	}
	if (tmp == checkbuf)	/* / has to exist */
		return 1;

	*tmp = '\0';

	return(gettype_stat(checkbuf, (struct stat *) 0) != -1);
}


int
makeroot(ptr, uid, gid, mode)
char *ptr;
int uid, gid, mode;
{
	char makebuf[LINELEN], *tmp;
	int usave;
	struct stat sbuf;


	strcpy(makebuf, ptr);
	if ((tmp = strrchr(makebuf, '/')) == NULL) {
		sprintf(errmsg, "makeroot can't find / in %s", ptr);
		do_gripe();
		return(-1);
	}
	*tmp = '\0';

	if (gettype_stat(makebuf, &sbuf) != -1)
		return(0);

	if (makeroot(makebuf, uid, gid, mode) != 0)
		return -1;

	if (verboseflag)
		printf("making root directory %s mode = %o\n", makebuf, mode);

	usave = umask(000);

	if ( mkdir(makebuf, (mode | 0111)) == -1) {
		sprintf(errmsg, "can't create directory %s", makebuf);
		do_gripe();
		umask(usave);
		return(-1);
	}
	umask(usave);

	if (uid != sbuf.st_uid || gid != sbuf.st_gid) {
		if (chown(makebuf, uid, gid) == -1) {
			sprintf(errmsg, "can't chown directory %s %d %d",
				makebuf, uid, gid);
			do_gripe();
		}
	}
#if 0
	/*
	 * If we chmod the directory first, then we can't write into it!
	 * Should save all directories created and modes, and do them
	 * all before quitting?
	 */
	if ( chmod(makebuf, (mode | 0111)) == -1) {
		sprintf(errmsg, "can't chmod directory %s %d %d",
			makebuf, uid, gid);
		do_gripe();
	}
#endif
	return 0;
}


removeit(ptr, type)
char *ptr;
int type;	/* file type	*/
{
	struct dirent *next;
	DIR * dirp;
	char rmbuf[LINELEN];

	if (verboseflag)
		printf("removing %s\n", ptr);
	switch (type) {
	case 'f' :
	case 'l' :
		return(unlink(ptr));
	default :
		sprintf(errmsg, "removeit got funny code %d for %s", type,
			ptr);
		return(-1);
		;
	case 'd' :
		/* Why not make life easier, and just system("rm -rf ...")? */
		if ((dirp = (DIR * )opendir(ptr)) == (DIR * )NULL) {
			sprintf(errmsg, "removeit: opendir(%s) failed", ptr);
			do_gripe();
			return(-1);
		}
		while ((next = readdir(dirp)) != NULL) {
			if (strcmp(next->d_name, ".") &&  strcmp(next->d_name,
			     "..")) {
				/* XXX check for buffer overflow */
				strcpy(rmbuf, ptr);
				strcat(rmbuf, "/");
				strcat(rmbuf, next->d_name);
				type = gettype_stat(rmbuf, (struct stat *) 0);
				if (removeit(rmbuf, type) == -1) {
					sprintf(errmsg,
						"removeit can't remove %s",
						rmbuf);
					do_gripe();
					return(-1);
				}
			}
		}
		closedir(dirp);
		if (rmdir(ptr) == -1) {
			sprintf(errmsg, "removeit can't remove directory %s",
				ptr);
			do_gripe();
			return(-1);
		}
		break;
	}
	return(0);
}


