/* Error routines and some miscellaneous functions */
#ifndef lint
static char rcsid[] = "$Id: misc.c,v 1.15 91/04/23 12:10:02 root Exp $";
#endif

#include "track.h"

static void
printmsg()
{
	fflush(stdout);	/* make sure any verbosity is flushed first */
	if (errno) {
		perror(errmsg);
		errno = 0;
	} else
		fprintf(stderr, "%s\n", errmsg);
	if (cur_ent >= 0 && cur_ent < entrycnt && entries[cur_ent].fromfile)
		fprintf(stderr, "\titem %s\n", entries[cur_ent].fromfile);
	fflush(stderr);
}


do_gripe()
{
	if (!quietflag)
		printmsg();
}


do_panic()
{
	printmsg();
	/* abort();*/
	cleanup(1);
}


char *
mycpy(to, tosz, from)
register char *to, *from;
int tosz;
{
	register char *end = to + tosz;

	while ((*from != '\0') && isprint(*from) && (!isspace(*from))) {
		if (to == end) {
			sprintf(errmsg,
			   "mycpy: destination buffer only %d, won't hold %s",
				tosz, from);
			do_panic();
			/*NOTREACHED*/
		}
		*to++ = *from++;
	}
	*to = '\0';
	
	return to;
}


skipword(theptr)
register char **theptr;
{
	register int c;

	while ((c = **theptr) != '\0' && isprint(c) && !isspace(c)) {
		(*theptr)++;
	}
}


skipwordspace(theptr)
char **theptr;
{
	register int c;
	register char *ptr = *theptr;

	while ((c = *ptr) != '\0' && isprint(c) && !isspace(c))
		ptr++;
	while (isspace(*ptr))
		ptr++;
	*theptr = ptr;
}


skipspace(theptr)
register char **theptr;
{

	while (isspace(**theptr)) {
		(*theptr)++;
	}
}


doreset()
{
	linebuf[0] = '\0';
	wordcnt = 0;
}


nofunny(ptr)
register char *ptr;
{
	for (; ; )
		switch (*ptr++) {
		case '*':
		case '[':
		case ']':
		case '?':
			return 0;
		case 0:
			return 1;
		}
}


/*
 * convert shell type regular expressions to egrep style form.  This doesn't
 * do a complete conversion. just a very specialized one.  The pattern is
 * anchored at the start, and either anchored at the end or followed by a
 * '/'.  '*' matches a run of any character but '/', '?' matches any single
 * character except '/'.  '.' matches the period literally.  Any other
 * character is passed straight through to regexp, so if regexp metacharacters
 * are used, weird things may happen. [] can probably be used safely as long
 * as there are no '*' or '?' in them.
 */
char *
regconv(from)
char *from;
{
	/* enough space for expanding a worst case pattern (all *) */
	char *tmp = emalloc(strlen(from) * 5 + 9);
	char *to;

	to = tmp;
	*to++ = '^';
	while (*from != '\0') {
		switch (*from) {
		case '.' :
			*to++ = '\\';
			*to++ = '.';
			from++;
			break;

		case '*' :
			*to++ = '[';
			*to++ = '^';
			*to++ = '/';
			*to++ = ']';
			*to++ = '*';
			from++;
			break;

		case '?' :
			*to++ = '[';
			*to++ = '^';
			*to++ = '/';
			*to++ = ']';
			from++;
			break;
		default:
			*to++ = *from++;
			break;
		}
	}
	*to++ = '(';
	*to++ = '/';
	*to++ = '.';
	*to++ = '*';
	*to++ = ')';
	*to++ = '?';
	*to++ = '$';
	*to++ = '\0';
	return(tmp);
}


/* convert shell-pattern to egrep and compile */
regexp *
regshcomp(pat)
char *pat;
{
	regexp *re = NULL;
	char *shre;

	if (!nofunny(pat)) {
		shre = regconv(pat);
		re = regcomp(shre);
		free(shre);
		if (re == NULL) {
			error("regcomp(%s) returned NULL", regconv(pat));
			/*NOTREACHED*/
		}
	}
	return re;
}



clear_ent()
{
	register int i;
	char ebuf[WORDLEN], *eptr;

	entries[entrycnt].followlink  = 0;
	entries[entrycnt].fromfile    = NULL;
	entries[entrycnt].tofile      = NULL;
	entries[entrycnt].cmpfile     = NULL;
	
	/* add global exceptions */
	eptr = g_except;
	skipspace(&eptr);
	for (i = 0; *eptr != '\0'; i++) {
		(void) mycpy(ebuf, sizeof ebuf, eptr);
		entries[entrycnt].exceptions[i] = strsave(ebuf);
		entries[entrycnt].re[i] = regshcomp(ebuf);
		skipwordspace(&eptr);
	}

	/* and clear the rest of the exceptions	*/
	for (; i <= WORDMAX; i++) {
		entries[entrycnt].exceptions[i] = 0;
		entries[entrycnt].re[i] = 0;
	}
	entries[entrycnt].cmdbuf      = 0;
}


parseinit()
{
	yyin = subfile;
	yyout = stderr;
	doreset();
}

/* for lex -- called on end of file. */
int
yywrap()
{
	return 1;
}

#ifndef CSRIMALLOC
char *
strsave(s)
char *s;
{
	return strcpy((char *) emalloc(strlen(s) + 1), s);
}

univptr_t
emalloc(len)
{
	char *p;

	if ((p = malloc((unsigned) len)) == NULL) {
		sprintf(errmsg, "couldn't allocate %d bytes", len);
		do_panic();
	}
	return p;
}

#endif
