/*
 * rips off the newline, if any at the end of theline, and appends the mtime
 * of the file named by the last word on the line.  If st is non-NULL, it is
 * assumed to be a pointer to a stat struct for the filename.  The first
 * character of theline is assumed to be one of 'l', 'd', or 'f'.
 */
#ifndef lint
static char rcsid[] = "$Id: stamp.c,v 1.11 91/04/23 12:10:09 root Exp $";
#endif

#include "track.h"
#include "pgcache.h"

#define NOREG "0"
#define NODIR "-1"
#define NOLNK "-2"

int
nstamp(ftype, fname, stp, buf, bufsz)
char ftype;
char *fname;
struct stat *stp;
char *buf;
int bufsz;
{
	struct passwd *pw;
	struct group *gr;
	struct stat sbuf;
	char *add;
	int addlen = 0;

	if (stp == NULL) {
		stp = &sbuf;
		/* we use the file type we are handed */
		(void) gettype_stat(fname, stp);
	}
	add = buf;
#define MINSTAMPLEN 64
	if (bufsz < MINSTAMPLEN) {
		/* won't hold directory stamp, maybe even filetime */
		sprintf(errmsg, "stamp: bufsz %d too small, min %d\n", bufsz,
			MINSTAMPLEN);
		do_panic();
		/* NOTREACHED */
	}
	switch (ftype) {
	case 'f':
		/*
		 * regular files add the time of last modification or 0 if
		 * non-existent.
		 */
		if ((stp->st_mode & S_IFMT) == S_IFREG) {
			sprintf(add, "%ld", (long)stp->st_mtime);
			addlen = strlen(add);
		} else {
			memcpy(add, NOREG, sizeof(NOREG));
			addlen = sizeof(NOREG) - 1;
		}
		break;
	case 'd':
		/*
		 * directories add the owner, group, and mode, -1 if
		 * non-existent or not a directory
		 */
		if ((stp->st_mode & S_IFMT) == S_IFDIR) {
			if ((pw = getpwuid_cache(stp->st_uid)) == NULL) {
				sprintf(errmsg, "getpwuid(%d) failed, file %s",
				 	stp->st_uid, fname);
				do_gripe();
			}
			if ((gr = getgrgid_cache(stp->st_gid)) == NULL) {
				sprintf(errmsg, "getgrgid(%d) failed, file %s",
					stp->st_gid, fname);
				do_gripe();
			}
			if (pw == NULL || gr == NULL)
				return 0;
			sprintf(add, "%s.%s.%05o", pw->pw_name,
				gr->gr_name, stp->st_mode & MODE_BITS);
			addlen = strlen(add);
		} else {
			memcpy(add, NODIR, sizeof(NODIR));
			addlen = sizeof(NODIR) - 1;
		}
		break;
	case 'l':
		/* symbolic links add evaluation of the link */
		if ((stp->st_mode & S_IFMT) == S_IFLNK &&
		    (addlen = readlink(fname, add, LINELEN)) >= 0) {
			if (addlen == LINELEN) {
				/*
				 * link could be too long -- readlink
				 * doesn't give an error if the buffer
				 * is too short for the full name, it
				 * just puts what it can in.  Blech!
				 */
				add[LINELEN-1] = '\0';
				sprintf(errmsg, "symlink %s too long -- %s",
					fname, add);
				do_gripe();
				return 0;
			}
			/* null terminate the link name */
			add[addlen++] = '\0';
		} else {
			memcpy(add, NOLNK, sizeof(NOLNK));
			addlen = sizeof(NOLNK) - 1;
		}
		break;
	default:
		sprintf(errmsg, "stamp: unknown filetype %d for %s",
			ftype, fname);
		do_panic();
		/* NOTREACHED */
	}
	return addlen;
}
