/*
 * Copyright (c) University of Toronto 1990.
 * All rights reserved.
 * Written by Mark Moraes.
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or of the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. The authors are not responsible for the consequences of use of this
 *    software, no matter how awful, even if they arise from flaws in it.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Since few users ever read sources,
 *    credits must appear in the documentation.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.  Since few users
 *    ever read sources, credits must appear in the documentation.
 *
 * 4. This notice may not be removed or altered.
 */

/*
 * error - print best error message possible and exit
 */
/* Mark Moraes, University of Toronto */
#include "stash.h"
#include <sys/types.h>
#include "libc.h"

static Stash *
newstashstruct(nstp)
Stash *nstp;
{
	Stash *stp = (Stash *) emalloc(sizeof(Stash));
	
	stp->next = nstp;
	stp->left = STASHSIZE - 1;
	stp->cp = stp->buf;
	stp->cp[stp->left] = '\0';
	return stp;
}
	
Stash *
StashCreate()
{
	return newstashstruct((Stash *) 0);
}

void
StashAdd(stpp, ptr, len)
Stash **stpp;
char *ptr;
int len;
{
	Stash *stp = *stpp;
	int n = stp->left - len;

	if (n >= 0) {
		memcpy(stp->cp, ptr, len);
		stp->cp[len] = '\0';
		stp->cp += len;
		stp->left -= len;
	} else {
		memcpy(stp->cp, ptr, stp->left);
		stp->cp[stp->left] = '\0';
		ptr += stp->left;
		len -= stp->left;
		/* just in case anyone tries to use it again */
		stp->cp = 0;
		stp->left = 0;
		*stpp = newstashstruct(stp);
		StashAdd(stpp, ptr, len);
	}
}

int
StashApply(stpp, proc, arg)
Stash **stpp;
int (* proc)();
char *arg;
{
	Stash *stp = *stpp;
	int len = STASHSIZE - stp->left - 1;
	int ret = 0;
	
	if (stp->next != 0)
		/*
		 * first apply to rest of list recursively so the
		 * segments are handed to proc in the right sequence
		 */
		ret = StashApply(&stp->next, proc, arg);

	/* then, if that succeeded, do this element */
	if (ret == 0 && len != 0)
		ret = (* proc)(stp->buf, len, arg);
	return ret;
}

void
StashFree(stpp)
Stash **stpp;
{
	Stash *stp = *stpp;

	if (stp->next != 0)
		/* first free rest of list recursively */
		StashFree(&stp->next);
	free((char *) stp);
	*stpp = 0;
}

#ifdef STASHTEST
#include <stdio.h>

char *progname = "st";

static int
printstash(ptr, len, arg)
char *ptr;
int len;
char *arg;
{
	if (arg != 0)
		fputs(ptr, (FILE *) arg);
	else
		fprintf(stderr, "\"%s\"\n", ptr);
	return 0;
}

int
main()
{
	char buf[128];
	Stash *stp;

	mal_debug(3);
	stp = StashCreate();
	while(fgets(buf, sizeof(buf), stdin) != NULL) {
		StashAdd(&stp, buf, strlen(buf));
		fprintf(stderr, "Stash is:\n");
		StashApply(&stp, printstash, (char *) 0);
		fprintf(stderr, "======\n\n");
	}
	StashApply(&stp, printstash, (char *) stdout);
	StashFree(&stp);
	fprintf(stderr, "stp = 0x%lx\n", (unsigned long) stp);
	return 0;
}
#endif
