/* Mark Moraes, University of Toronto */
/* 
 * A Stash is a string, stored in a list of non-contiguous segments.
 * It's innards may not be directly accessed -- text may be appended to
 * it via StashAdd(), or a procedure can be applied to each of the
 * non-contiguous segments in sequence using StashApply().
 */

#ifdef STASHDEBUG
# define STASHSIZE 10
#else
# define STASHSIZE 1000
#endif

typedef struct stashstruct {
	struct stashstruct *next;
	char buf[STASHSIZE];
	char *cp;
	int left;
} Stash;

/* Allocates a new stash and returns a pointer to it. */
extern Stash *StashCreate(/* void */);

/* Adds len characters from the buffer ptr to the end of the stash */
extern void StashAdd(/* Stash **stpp, char *ptr, int len */);

/* 
 * This takes a function proc(), and applies it to each segment of the
 * stash in sequence.  proc() is called as proc(char *ptr, int len, char
 * *arg) where ptr is a pointer to a segment of len characters.  The
 * segment is nul-terminated for convenience.  arg is the argument
 * passed into proc(), also for convenience (eg.  In passing a file
 * pointer) If proc() returns non-zero, then the apply is stopped
 * prematurely at that segment and the value returned by proc() is
 * returned.  StashApply() returns 0 on successful completion (no proc()
 * returned 0)
 */
extern int StashApply(/* Stash **stpp, int (* proc)(), char *arg */);

/* Frees a stash.  The stash may not be referred to again */
extern void StashFree(/* Stash **stpp */);
