#ifndef lint
static char rcsid[] = "$Id: track.c,v 1.54 91/08/23 02:11:21 ken Exp $";
#endif

#include "track.h"
#include <fcntl.h>
#include <sys/file.h>
#include "lock.h"
#include <sys/errno.h>

#ifndef MAXHOSTNAMELEN
# define MAXHOSTNAMELEN 64
#endif

/* Not many letters left... Grab 'em while they last! */
#define USAGE \
"Usage: %s [-d localdir] [-D] [-f] [-g subscriber] [-l librarian]\n\
       [-m admin] [-n] [-o] [-O] [-p] [-q] [-r remotedir] [-s] [-t] [-u]\n\
       [-U user] [-v] [-w]\n"

char *admin = DEF_ADM;			/* person(s) to mail errors to */
char *rootdir = DEF_ROOT;	/* default track working directory */
char *rrootdir = 0;		/* remote track working directory */
char g_except[LINELEN] = DEF_EXCEPT;	/* XXX */

char errmsg[3 * LINELEN];	/* enough to hold 2 lines plus error msg */
char *progname = "track";

int via = VIA_TCP;	/* mode of communications, VIA_TCP, VIA_RSH */

int child_id;		/* calltrack or rsh process id */
FILE *statfile;

int writeflag = 0;	/* if set, output a statfile rather than
			   pulling files */
int parseflag = 0;	/* if set, just parse the subscription list */
int forceflag = 0;	/* if set, fix mismatched file types (symlink vs
			   directory/file, etc) */
int verboseflag = 0;	/* if set, print trace information */
int dirflag = 0;	/* if set, create directories if necessary */
int nopullflag = 0;	/* if set, find out the differences,
                           but don't pull anything */
int quietflag = 0;	/* if set, don't print non-fatal error messages */
int interactive = 1;	/* if set, don't send errors via mail, print them */
int uflag = NO_CLOBBER;	/* if set, copy a older file on top of a newer one */
int getflag = 0;	/* if set, should copy out the appropriate files */
int listflag = 0;	/* if set, should copy out the appropriate stafile */
int optimizeflag = 0;	/* if set, only do something if statfiles differ */

char *subname = NULL;		/* subscription file  */
FILE *subfile;			/* pointer to the subscription list */

char myname[MAXHOSTNAMELEN];	/* my host name */
char *user = "";		/* user to use on distant machine */
char *subscriber;		/* track client name for -g */
char *librarian = NULL;		/* distant machine name */
int sep_char = DEF_SEP_CHAR;	/* field separator in headers */

Entry entries[ENTRYMAX];	/* stores the slist.  should be dynamic. */
int entrycnt = 0;		/* actual number of entries <= ENTRYMAX */

/*
 * cur_ent is a global pointing at the current slist entry we're working
 * on.  It's used in far too many places -- be careful if you change it.
 */
int cur_ent = 0;	/* XXX */

int gargc;
char **gargv;

extern long gettime();
static int ecmp();

int
main(argc, argv)
int argc;
char **argv;
{
	char subtmp[LINELEN];
	int c;

	errno = 0;
#ifdef DEBUG
	openlog("track test", LOG_PID);
#endif

	errmsg[0] = '\0';

	gargc = argc; 
	gargv = argv;
	gethostname(myname, sizeof myname);

	umask(022);	/* set default umask for daemons */

#ifdef CSRIMALLOC
	mal_debug(3);
	mal_trace(1);
#endif
	while ((c = getopt(argc, argv, "d:Dfg:l:m:noOpqr:stuU:vw")) != EOF) {
		switch (c) {
		case 'd':
			rootdir = optarg;
			break;
		case 'D':
			dirflag = 1;
			break;
		case 'f':
			forceflag = 1;
			break;
		case 'g':
			interactive = 0;
			getflag = 1;
			subscriber = optarg;
			break;
		case 'l':
			librarian = optarg;
			break;
		case 'm':
			interactive = 0;
			admin = optarg;
			break;
		case 'n':
			nopullflag = 1;
			break;
		case 'o':	/* obsolete */
			break;
		case 'O':
			optimizeflag++;
			break;
		case 'p':
			parseflag = 1;
			break;
		case 'q':
			quietflag = 1;
			break;
		case 'r':
			rrootdir = optarg;
			break;
		case 's':
			listflag = 1;
			break;
		case 't':
			via = VIA_RSH;
			break;
		case 'u':
			uflag = DO_CLOBBER;
			break;
		case 'U':
			user = optarg;
			break;
		case 'v':
			verboseflag = 1;
			break;
		case 'w':
			writeflag = 1;
			break;
		case '?':
			fprintf(stderr, USAGE, progname);
			exit(1);
		}
	}
	if (optind == argc - 1) {
		subname = argv[optind];
	}
	if (!interactive) {
		setuperr();
	}
	if (subname == NULL) {
		strcpy(errmsg, "Must have an slist");
		do_panic();
		/* NOTREACHED */
	}
	/*
	 * by default, use the same directory on remote machines as we are
	 * using on this machine
	 */

	if (uflag && optimizeflag) {
		strcpy(errmsg, "cannot specify both -u and -O.  -O ignored");
		do_gripe();
		optimizeflag = 0;
	}
	if (rrootdir == 0) {
		rrootdir = rootdir;
	}

	if (chdir(rootdir) == -1) {
		sprintf(errmsg, "can't chdir to %s", rootdir);
		do_panic();
	}

	if (verboseflag)
		lockdebug(stdout);
		
	if (listflag) {
		sendlist();
		cleanup(0);
	}

	sprintf(subtmp, "slists/%s", subname);
	if ((subfile = fopen(subtmp, "r")) == NULL) {
		sprintf(errmsg, "Can't open %s", subtmp);
		do_panic();
	}

	/*
	 redirect yacc/lex i/o
	*/
	parseinit();
	if (yyparse()) {
		strcpy(errmsg, "parser error");
		do_panic();
	}

	qsort((char *) entries, entrycnt, sizeof(Entry), ecmp);

	if (verboseflag) {
		printf("parse worked\n");
	}

	if (parseflag) {
		justshow(stdout);
		cleanup(0);
	}

	if (writeflag) {
		writenames();
		cleanup(0);
	}

	if (getflag) {
		sendfiles();
		cleanup(0);
	}

	if (librarian == NULL) {
		strcpy(errmsg, "Must specify a librarian (-l)");
		do_panic();
		/* NOTREACHED */
	}

	if (nopullflag)	 {
		readnames(1);
		cleanup(0);
	}

	readnames(0);

	cleanup(0);
	return 0;	/* keeps gcc quiet, really NOTREACHED */
}


static int
has_perm()
{
#define MAXFIELDS 64
	FILE * permfile;
	char permbuf[LINELEN], *word[MAXFIELDS];
	int i, n, found, trip;

	/*
	 * if the permissions file does not exist, return the default
	 * permission
	 */
	if ((permfile = fopen("permissions", "r")) == NULL) {
		return(DEF_PERM);
	}

	/*
	 * The permissions file looks like
	 *	thing s1 s2 ...
	 * where thing is a host name or a netgroup name,
	 * and the s* are subscription list names.
	 * The special string ANY matches anything.
	 */
	found = 0;
	while (fgets(permbuf, sizeof permbuf, permfile) != NULL) {
		if (permbuf[0] == '#' ||
		    (n = split(permbuf, word, MAXFIELDS, " \t\n")) < 2)
			continue;
#ifndef INNETGR
# define innetgr(a,b,c,d) 0
#endif
		if (strcmp(word[0], "ANY") != 0 &&
		    strcmp(word[0], subscriber) != 0 &&
		    !innetgr(word[0], subscriber, (char *)NULL, (char *)NULL))
			continue;
		/* First field matches.  Check list names. */
		for (i = 1; i < n; i++)
			if (strcmp(word[i], subname) == 0 ||
			    strcmp(word[i], "ANY") == 0) {
				found++;
				break;
			}
	}
	fclose(permfile);
	return found;
}


sendlist()
{
	char statname[LINELEN];
	log("requesting list");
	if (!has_perm()) {
		log("no permission");
		copyerror("permission denied");
		sprintf(errmsg, "%s does not have permission to read %s",
			subscriber, subname);
		do_panic();
	}
	inpipe = 0;	/*
			 * don't know why I open this, i never read it when
			 * sending a statfile
			 */
	outpipe = 1;
	sprintf(statname, "%s/stats/%s.%s", rootdir, subname, myname);
	writefile(statname, statname);
}


/* 
 * anygood checks if the filename 'ptr' matches anything in the slist.
 * NOTE: this modifies cur_ent
 */
int
anygood(ptr)
char *ptr;
{
	Entry *ep;
	int rcmp;
	
	for (cur_ent = 0; cur_ent < entrycnt; cur_ent++) {
		ep = &entries[cur_ent];
		rcmp = headmatch(ptr, ep->fromfile, ep->fromlen);
		if (rcmp < 0) {
			/* not in slist (since slist is sorted) */
			return 0;
		} else if (rcmp == 0) {
			/* matches the slist -- check for exceptions */
			if (ngoodname(ptr + ep->fromlen, cur_ent))
				/* yup, this one's ok */
				return(1);
			/* nope, an exception, keep going */
		}
	}
	return(0);
}


sendfiles()
{
	char lbuf[LINELEN], physname[LINELEN];
	int hlen;

	log("requesting files");
	if (!has_perm()) {
		log("permission denied");
		copyerror("permission denied");
		sprintf(errmsg, "%s does not have permission to read %s",
			subscriber, subname);
		do_panic();
		/* NOTREACHED */
	}

	inpipe = 0;
	outpipe = 1;
	while ((hlen = get_head(lbuf, sizeof lbuf)) != 0) {
		if (anygood(lbuf)) {
			memcpy(physname, lbuf, hlen + 1);
			writefile(lbuf, physname);
			log(lbuf);
		} else {
			sprintf(errmsg, "%s not in slist", lbuf);
			copyerror(errmsg);
			log(errmsg);
		}
	}
}


void
fprintcmd(cmd, fp)
char *cmd;
FILE *fp;
{
	char *cp = cmd + strlen(cmd) -1;

	while (cp >= cmd && *cp == '\n')
		*cp-- = '\0';
	cp = cmd;
	while (*cp != '\0') {
		if (*cp == '\n')
			fputc('\\', fp);
		fputc(*cp, fp);
		cp++;
	}
	fputs("\\\n\n", fp);
}


static int 
ecmp(p1, p2)
char *p1;
char *p2;
{
	return strcmp(((Entry * )p1)->fromfile, ((Entry * )p2)->fromfile);
}


justshow(fp)
FILE *fp;
{
	int i, j;
	Entry * ep;

	fputs("#\n# source:comparison:destination:exceptions:script\n#\n",
	     	      fp);
	for (i = 0; i < entrycnt; i++) {
		ep = &entries[i];
		fprintf(fp, "%s%s:%s:%s:", ep->followlink ? "!" : "",
			ep->fromfile,
			ep->cmpfile == NULL ? "" :  ep->cmpfile,
			ep->tofile == NULL ? "" :  ep->tofile);
		if (ep->exceptions[0] == 0)
			fputc(':', fp);
		else {
			fputc('\n', fp);
			for (j = 0; ep->exceptions[j] != 0; j++) {
				fprintf(fp, "\t%s\n", ep->exceptions[j]);
			}
			fputs("\t:", fp);
		}
		if (ep->cmdbuf && ep->cmdbuf[0] != '\0') {
			fprintcmd(ep->cmdbuf, fp);
		} else {
			fputc('\n', fp);
		}
	}
}

/*
 * redirect standard error output to mail to the administrator. Use nullmail so
 * that null messages won't get sent.
 */
setuperr()
{
#define MAILCMD "/bin/nullmail"
	char *cmd = emalloc(strlen(rootdir) + sizeof MAILCMD);
	char *args[5];
	int pid, fds[2];

	strcpy(cmd, rootdir);
	strcat(cmd, MAILCMD);
	if (access(cmd, 01 | 04) < 0)
		goto err;

	args[0] = cmd;
	args[1] = subname ? subname : "";
	args[2] = admin ? admin : DEF_ADM;
	args[3] = subscriber;
	args[4] = NULL;

	/* start a process that will send mail to the adminstrator */
	if (pipe(fds) < 0)
		goto err;
	if ((pid = fork()) < 0)
		goto err;
	else if (pid == 0) {	/* child */
		if (dup2(fds[0], 0) < 0)
			goto err;
		(void) close(fds[0]);
		(void) close(fds[1]);
		execv(cmd, args);
		goto err;
	}
	/* parent -- now connect stderr to the pipe */
	(void) close(fds[0]);
	if (dup2(fds[1], 2) < 0)
		goto err;
	(void) close(fds[1]);
	return 0;
err:
	sprintf(errmsg,
"echo HELP automatic software distribution --%s cant execute nullmail cmd\
> /dev/console",
			gargv[0]);
	(void) system(errmsg);
	exit(1);
	/* NOTREACHED */
}

static char lockfile[LINELEN];
static int locked;

setlock()
{
	sprintf(lockfile, "%s/locks/track.%s", rootdir, subname);
	if (lockcreate(lockfile, 15, (unsigned) 1) != 0) {
		sprintf(errmsg, "lock set on %s; quitting", subname);
		errno = 0;
		do_gripe();
		cleanup(1);
		/* NOTREACHED */
	}
	locked++;
}


clearlock()
{
	if (locked) {
		(void) lockfree(lockfile);
		locked = 0;
	}
}

static FILE *stampfile;		/* used to write out the statfile */

/*
 * Uses the slist to scan the filesystem and make up the statfile.
 * NOTE:  modifies cur_ent
 */
writenames()
{
	char lbuf[LINELEN], outname[LINELEN];
	char oldoutname[LINELEN];
	int type;
	struct stat st;
	Entry *ep;
	time_t start_time[2];

	setlock();

	sprintf(oldoutname, "stats/%s.%s", subname, myname);
	sprintf(outname, "%s.tmp", oldoutname);

	if ((stampfile = fopen(outname, "w")) == NULL) {
		sprintf(errmsg, "Can't write new stat file %s", outname);
		do_panic();
		/* NOTREACHED */
	}
	/* saved start time for the optimize flag */
	start_time[0] = time((time_t *) 0);
	for (cur_ent = 0; cur_ent < entrycnt; cur_ent++) {
		ep = &entries[cur_ent];
		switch (type = gettype_stat(ep->fromfile, &st)) {
		case 'd' :
			dofinds();
			break;
		case 'l':
		case 'f':
			/*
			 * no need to check for exceptions since this entry is
			 * either a file or a symbolic link and thus it won't
			 * result it is the root of a trivial subtree
			 */
			if (!nstamp(type, ep->fromfile, &st,
				    lbuf, sizeof(lbuf)))
				break;
			putc(type, stampfile);
			fputs(ep->fromfile, stampfile);
			putc(' ', stampfile);
			fputs(lbuf, stampfile);
			putc('\n', stampfile);
			break;
		default:
			sprintf(errmsg, "got %d from gettype of %s", type,
				ep->fromfile);
			do_gripe();
			break;
		}
	}
	if (fclose(stampfile) < 0) {
		sprintf(errmsg, "couldn't close new stat file %s", outname);
		do_panic();
		/* NOTREACHED */
	}
	start_time[1] = time((time_t *) 0);
	/*
	 * fake the mod time on the file to be the time we STARTED writing the
	 * file.  This guarantees that a client will not miss any changes that
	 * took place while the file was being written.
	 */
	if (utime(outname, start_time) < 0) {
		sprintf(errmsg, "couldn't set utime on %s", outname);
		do_panic();
		/* NOTREACHED */
	}
	/* 
	 * now compare the newly created statfile with the old one if
	 * they differ, replace the old with the new
	 */
	if (gettype_stat(oldoutname, (struct stat *) 0) != -1) {
		switch (cmp(outname, oldoutname)) {
		case 0:
			/* no difference, so just remove the new one */
			if (unlink(outname) == -1) {
				sprintf(errmsg, "can't remove %s", outname);
				do_panic();
				/* NOTREACHED */
			}
			break;
		case 1:
			/*
			 * something has changed, replace the old with the new
			 */
			if (rename(outname, oldoutname) == -1) {
				sprintf(errmsg, "can't rename %s to %s",
					outname, oldoutname);
				do_panic();
			}
			break;

		case -1:
		case 2:
		default:
			sprintf(errmsg, "can't compare %s and %s", outname,
				oldoutname);
			do_panic();
			/* NOTREACHED */
		}
	} else {
		if (rename(outname, oldoutname) == -1) {
			sprintf(errmsg, "can't rename %s to %s", outname,
				oldoutname);
			do_panic();
		}
	}
}


static char *cur_fromfile;

findfunc(s, st)
char *s;
struct stat *st;
{
	char buf[LINELEN];
	int c;

	switch (st->st_mode & S_IFMT) {
	case S_IFREG: 
		c = 'f'; 
		break;
	case S_IFDIR: 
		c = 'd'; 
		break;
	case S_IFLNK: 
		c = 'l'; 
		break;
	default:
		/* prune on special file */
		return 1;		/* XXX */
	}
	if (goodname(s, cur_fromfile, cur_ent)) {
		/*
		 * sep_char in stats files will cause parsing errors at the
		 * other end.  There should be some kind of escape mechanism to
		 * handle this.  For now, just skip them.
		 */
		if (strchr(s, sep_char)) {
			sprintf(errmsg,
				"Can't handle '%c' in \"%s\"; skipping",
			     	sep_char, s);
			do_gripe();
			return 1;
		}
		if (!nstamp(c, s, st, buf, sizeof buf))
			return 1;
		putc(c, stampfile);
		fputs(s, stampfile);
		putc(' ', stampfile);
		fputs(buf, stampfile);
		putc('\n', stampfile);
		return 0;
	}
	return 1;	/* prune */
}


/* findfunc only called from child find(), and it doesn't change cur_ent */
dofinds() 
{
	cur_fromfile = entries[cur_ent].fromfile;
	find(cur_fromfile, findfunc);
	cur_fromfile = NULL;
}

/* 
 * are the two files in a stamp record different? For symlinks and
 * directories, any difference in the stamp is significant, for files,
 * we normally bring over only newer files.  If DO_CLOBBER is set, then
 * we bring over files that have a different stamp even if they are newer
 */
int
isdiff(ftype, p1, p2)
char ftype;
register char *p1;	/* remote stamp */
register char *p2;	/* local stamp */
{
	extern int numcmp();
	char lbuf[LINELEN];

	if (*p2 == '-')		/* local side doesn't have it */
		return 1;
	if (ftype == 'f' && uflag != DO_CLOBBER) {
		return numcmp(p1, p2) != 0;
	} else if (ftype == 'l') {
		/* we have to remap remote symlink name before comparing */
		int len = strlen(p1) + 1;

		if (len > sizeof lbuf) {
			sprintf(errmsg, "isdiff: need %d bytes, have only %d bytes\n",
				len, sizeof lbuf);
			do_panic();
		}
		(void) memcpy(lbuf, p1, len);
		(void) remapname(&entries[cur_ent], lbuf, sizeof lbuf, len);
		p1 = lbuf;
	}
	return !STREQ(p1, p2);
}


doconnect()
{
	static int have_conn = 0;

	if (!have_conn) {
		if (outpipe >= 0) {
			sprintf(errmsg, "doconnect -- already connected");
			do_panic();
		}
		startget(uflag, NO_LIST);
		have_conn = 1;
	}
}


disconnect()
{
	if (outpipe >= 0)
		stopget();
}


/*
 * compares the time on the remote statsfile with the time on the last local
 * copy of the remote statsfile (if any).  If the local copy is newer, then
 * get_stat returns 0, otherwise, it gets the remote statsfile over and
 * returns 1; statfile is a FILE pointer to the local copy of the statsfile.
 */
int
get_stat(tp)
long *tp;
{
	char statname[LINELEN], rstatname[LINELEN], hbuf[LINELEN];
	static stat_changed;
	struct stat st1;
	long remtime;
	int hlen;

	/* get ready to snarf the statfile from librarian */
	sprintf(statname, "%s/stats/%s.%s", rootdir, subname,
		librarian);
	sprintf(rstatname, "%s/stats/%s.%s", rrootdir, subname,
		librarian);

	/* need to the to detect if the statfile has changed and for -O */
	if (stat(statname, &st1) == -1) {
		st1.st_mtime = 0;
		*tp = 0;
	} else {
		*tp = st1.st_mtime;
	}
	if (verboseflag) {
		printf("remote working statfile='%s'\n", rstatname);
		printf("local working statfile='%s'\n", statname);
	}
	startget(DO_CLOBBER, IS_LIST);
	hlen = get_head(hbuf, sizeof hbuf);
	/*
	 * check for some bogus stuff ie "connection timed out" as the
	 * first line of input
	 */
	if (strncmp(hbuf, rstatname, strlen(rstatname))) {
		int childstatus;
		sleep(1);	/* give calltrack a chance to die */
		childstatus = reapchild();	/* get calltrack exit status */
		if (quietflag && childstatus != -1)
			/*
			 * calltrack is dead.
			 * Quietly die if the librarian is down
			 * or isn't listening, to avoid bothersome mail
			 * when a server is down. 
			 * (calltrack's exit value is from <errno.h>).
			 */
			switch ((childstatus>>8)&0xff) {
			case ECONNREFUSED:
			case ECONNRESET:
			case ETIMEDOUT:
			case EINTR:	/* ? */
				cleanup((childstatus>>8)&0xff);
			}
		sprintf(errmsg,
			"statfile header did not start with %s\n\t%s",
			rstatname, hbuf);
		/* Copy errno from calltrack's exit status */
		if (childstatus != -1 && ((childstatus>>8)&0xff) != 0)
			errno = (childstatus>>8) & 0xff;
		do_panic();
		/* NOTREACHED */
	}
	remtime = gettime(hbuf);
	if (verboseflag) {
		printf("local time = %ld, remote time = %ld\n",
		       (long) st1.st_mtime, remtime);
	}
	if (st1.st_mtime >= remtime) {
		skipbytes(inpipe, getsize(hbuf));
		stopget();
		if ((statfile = fopen(statname, "r")) == NULL) {
			sprintf(errmsg, "can't open statfile %s",
				statname);
			do_panic();
			/* NOTREACHED */
		}
		stat_changed = 0;
		return stat_changed;
	}
	if (get_body(hbuf, statname) != 0) {
		sprintf(errmsg, "couldn't get statfile %s", statname);
		do_panic();
		/*NOTREACHED*/
	}
	stopget();
	if ((statfile = fopen(statname, "r")) == NULL) {
		sprintf(errmsg, "can't open statfile %s", statname);
		do_panic();
		/* NOTREACHED */
	}
	stat_changed = 1;
	return stat_changed;
}

#ifdef HAVE_BCOPY
# define memmove(dst, src, len) bcopy((src), (dst), (len))
#endif

int
presub(line, linelen, linesz, old, olen, new, nlen)
char *line, *old, *new;
int linelen, olen, nlen;
{
	register int delta = nlen - olen;
	register int newlen = linelen + delta;

	if (memcmp(line, old, olen) != 0)
		return linelen;
	if (newlen + 1 > linesz) {
		sprintf(errmsg, "presub: need %d bytes, have only %d bytes",
			newlen + 1, linesz);
		do_panic();
	}
	memmove(line + nlen, line + olen, linelen - olen);
	memcpy(line, new, nlen);
	line[newlen] = '\0';
	return newlen;
}

/*
 * substitutes the current entry's cmpfile in place of fromfile in src,
 * placing the result in dest
 */
static int
getcmp(src, dest, destsz)
char *src, *dest;
int destsz;
{
	int len = entries[cur_ent].cmplen;
	char *start = dest;
	
	if (destsz <= len) {
		sprintf(errmsg, "getcmp:  buffer too short (%d)", destsz);
		do_panic();
		/* NOTREACHED*/
	}
	memcpy(dest, entries[cur_ent].cmpfile, len);
	dest += len;
	destsz -= len;
	len = entries[cur_ent].fromlen;
	return mycpy(dest, destsz, src + len) - start;
}


/*ARGSUSED*/
static int 
printstash(ptr, len, arg)
char *ptr;
int len;
char *arg;
{
	fputs(ptr, (FILE * ) arg);
	return 0;
}


runcmd(ep, notreally)
Entry *ep;
int notreally;
{
	FILE * cmdpipe;
	int tmpstat;

	if (ep->cmdbuf == NULL || ep->cmdbuf[0] == '\0' || ep->stash == NULL)
		return;
	if (verboseflag || notreally)
		printf("runcmd(%s), command: %s\n", ep->fromfile, ep->cmdbuf);
	if (! notreally) {
		if ((cmdpipe = popen(ep->cmdbuf, "w")) == NULL) {
			sprintf(errmsg, "can't execute command %s",
				ep->cmdbuf);
			do_panic();
		}
		StashApply(&ep->stash, printstash, (char *) cmdpipe);
		if ((tmpstat = pclose(cmdpipe)) != 0) {
			sprintf(errmsg, "got exit status %d from command %s",
				tmpstat, ep->cmdbuf);
			do_gripe();
		}
	}
	StashFree(&ep->stash);
	ep->stash = 0;
}

stashline(ep, lbufp)
Entry *ep;
char *lbufp;
{
	if (ep->cmdbuf && ep->cmdbuf[0]) {
		if (ep->stash == 0)
			ep->stash = StashCreate();
		StashAdd(&ep->stash, lbufp, strlen(lbufp));
	}
}

/* 
 * reads the statfile, finds files that match something on the slist,
 * checks if we need those files, and gets those files.  NOTE: This
 * modifies cur_ent.
 */
readnames(notreally)
int notreally;
{
	char lbuf[LINELEN];
	char *lbufp = lbuf;
	register int i;
	long tstamp = 0, mtime;

	setlock();
	if (get_stat(&tstamp) == 0 && optimizeflag) {
		if (verboseflag)
			printf("remote statsfile hasn't changed\n");
		disconnect();
		return;
	}

	/*
	 * This is the loop where client track spends almost all its time.
	 * Each line in the statsfile is compared to the slist, and if it
	 * matches an slist entry, the file is brought over as specified by
	 * the entry.  If the slist entry specifies a command, we want to run
	 * the command once, when all files matching it have been brought
	 * over.  So we stash the filenames as they are brought over for each
	 * slist entry with a command, and run the commands on each slist
	 * entry when we're all done, feeding the names of the changed files
	 * on stdin to the command.  We need this since many slist entries
	 * could be matched by a single statfile line.  So we've replaced
	 * holding the entire statfile in memory by holding the names of all
	 * files transferred in memory (worst case) which COULD be much
	 * worse.  Typical case behaviour is MUCH better, though.
	 */
	i = 0;
	while (fgets(lbuf, sizeof lbuf, statfile) != NULL) {
		char *fromfile;
		int fromlen, rcmp, ftype, nf;
		char cmpbuf[LINELEN], stampbuf[LINELEN];
#define ENFIELDS 3
		char *fields[ENFIELDS];
		char *cmpptr;
		Entry *ep;

		i++;
		lbufp = lbuf;
		ftype = *lbufp++;
		if ((nf = split(lbufp, fields, ENFIELDS, " \t\n")) != ENFIELDS) {
			sprintf(errmsg, "%d fields in line %d, expected %d",
				nf, i, ENFIELDS);
			do_gripe();
			continue;
		}
		/* fields[0] is the filename, fields[1] is the stamp */
		if (optimizeflag && ftype == 'f') {
			mtime = atol(fields[1]);
			if (mtime <= tstamp) {
				if (verboseflag)
					printf("skipping %s\n", fields[0]);
				continue;
			}
		}
		for (cur_ent = 0; cur_ent < entrycnt; cur_ent++) {
			ep = &entries[cur_ent];
			fromfile = ep->fromfile;
			fromlen = ep->fromlen;
			rcmp = headmatch(fields[0], fromfile, fromlen);
			if (rcmp > 0) {
				continue;
			} else if (rcmp < 0) {
				break;
			}
			if (!ngoodname(fields[0] + fromlen, cur_ent))
				continue;
			if (ep->cmpfile) {
				getcmp(fields[0], cmpbuf, sizeof cmpbuf);
				cmpptr = cmpbuf;
			} else {
				cmpptr = fields[0];
			}
			if (nstamp(ftype, cmpptr, (struct stat *)0,
				    stampbuf, sizeof stampbuf) &&
			    isdiff(ftype, fields[1], stampbuf)) {
				char hbuf[LINELEN];
				int hlen = strlen(fields[0]);

				memcpy(hbuf, fields[0], hlen + 1);
				(void) remapname(ep, hbuf, sizeof hbuf, hlen);
				if (notreally) {
					puts(hbuf);
					stashline(ep, hbuf);
					stashline(ep, "\n");
				} else {
					if (dofile(ep, fields[0], hlen, hbuf)
					    == 0) {
						stashline(ep, hbuf);
						stashline(ep, "\n");
					}
				}
			}
		}
	}
	for (cur_ent = 0; cur_ent < entrycnt; cur_ent++)
		runcmd(&entries[cur_ent], notreally);
	disconnect();
}

int
remapname(ep, hbuf, hsz, hlen)
Entry *ep;
char *hbuf;
int hsz, hlen;
{
	int newlen = hlen;
	
	if (ep->tofile)
		newlen = presub(hbuf, hlen, hsz, ep->fromfile,
		       ep->fromlen, ep->tofile, ep->tolen);
	else if (ep->cmpfile)
		newlen = presub(hbuf, hlen, hsz, ep->fromfile,
		       ep->fromlen, ep->cmpfile, ep->cmplen);
	return newlen;
}

int
dofile(ep, fname, flen, toname)
Entry *ep;
char *fname, *toname;
int flen;
{
	char hbuf[LINELEN], *hptr;
	int status = 0, hlen;

	doconnect();
	if (verboseflag)
		printf(">> %s\n", fname);
	pwrite(outpipe, fname, strlen(fname));
	pwrite(outpipe, "\n", 1);		/* and add a newline */
	hbuf[0] = '\0';
	if ((hlen = get_head(hbuf, sizeof hbuf)) == 0) {
		/*
		 * This will happen if calltrack goes away for whatever
		 * reason.  Or we might get a SIGPIPE from above first.  Sigh.
		 */
		sprintf(errmsg, "can't get_head for %s, got\n\t%s", fname,
			hbuf);
		do_gripe();
		return -1;
	}
	if (verboseflag)
		printf("<< %s\n", hbuf);

	if (hlen == 0 || hbuf[0] == '\0' || hbuf[0] == sep_char) {
		sprintf(errmsg, "trackd error getting file %s: %s", fname,
			hbuf);
		do_gripe();
		return -1;
	}
	if (memcmp(hbuf, fname, flen) == 0) {
		status = get_body(hbuf, toname);
	} else {
		sprintf(errmsg, "asked librarian for %s, got: %s\n",
			fname, hbuf);
		do_gripe();
		flush_body(hbuf);
		return -1;
	}
	return status;
}


log(ptr)
char *ptr;
{
	static FILE *logfile = NULL;
	static int nologfile;
	char namebuf[LINELEN];
	char timestring[LINELEN];
	char tstring[LINELEN];
	long timebuf;

	if (logfile == NULL) {
		if (nologfile)
			return;
		sprintf(namebuf, "%s/logfiles/lib.log", rootdir);
		if (gettype_stat(namebuf, (struct stat *) 0) == -1) {
			nologfile++;	/* so we won't keep stat()ing it */
			return;
		}
		if ((logfile = fopen(namebuf, "a")) == NULL) {
			sprintf(errmsg, "can't open log file %s", namebuf);
			do_panic();
			/* NOTREACHED */
		}
#ifdef _IOLBF
		setvbuf(logfile, NULL, _IOLBF, 0);
#else
		setlinebuf(logfile);
#endif
	}
	if (via == VIA_TCP) {
		strcpy(tstring, "-t");
	} else {
		strcpy(tstring, "  ");
	}
	timebuf = time((time_t *) 0);
	strcpy(timestring, ctime(&timebuf));
	*(timestring + strlen(timestring) - 1) = '\0';	/* wipe out newline */
	fprintf(logfile, "%s %s %s %s %s\n", timestring, subscriber, tstring,
		subname, ptr);
}


cleanup(err)
int err;
{
	clearlock();
	chdir("/tmp");	/* dump profiling data in a known place */
	exit(err);
}


/*
 * Same as "cmp -s file1 file2".
 */
cmp(file1, file2)
char *file1, *file2;
{
	register FILE *f1, *f2;
	register int n1, n2;
	char b1[8*1024], b2[8*1024];
	struct stat s1, s2;
	int status = 0;

	if ((f1 = fopen(file1, "r")) == NULL)
		return 2;
	if ((f2 = fopen(file2, "r")) == NULL) {
		fclose(f1);
		return 2;
	}
	if (fstat(fileno(f1), &s1) < 0 || fstat(fileno(f2), &s2) < 0) {
		status = 2;
		goto done;
	}
	if (s1.st_size != s2.st_size) {
		status = 1;
		goto done;
	}
	do {
		n1 = read(fileno(f1), b1, sizeof b1);
		n2 = read(fileno(f2), b2, sizeof b2);
		if (n1 != n2 || n1 < 0 || memcmp(b1, b2, n1) != 0) {
			status = 1;
			break;
		}
	} while (n1 > 0);
done:
	fclose(f1);
	fclose(f2);
	return status;
}


pwrite(fd, p, n)
char *p;
{
	int cc;
	char *p0 = p;

	while (n > 0) {
		cc = write(fd, p, n);
		if (cc <= 0) {
			perror("track: write (on pipe)");
			if (p == p0) 
				return -1;
			break;
		}
		n -= cc;
		p += cc;
	}
	return p - p0;
}


