
/*
**	track daemon
**	based on the protocol used for rsh
**
*/

#ifndef lint
static char rcsid[] = "$Id: trackd.c,v 1.21 91/04/23 12:10:16 root Exp $";
#endif


#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#ifdef HAVE_SYS_SELECT	/* AIX on RS/6000 */
# include <sys/select.h>
#endif

#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <netdb.h>
#include <syslog.h>

#include <ctype.h>
#define DAEMON
#include "track.h"
#include "sig.h"

#ifndef	FD_ZERO
#include	"fd-set.h"
#endif

#ifndef DEBUG
#define TRACKSERVICE	"track"
#define ERROR_EXIT(x)	exit(x)
#else
#define TRACKSERVICE	"tracktest"
#define ERROR_EXIT(x)	abort()
#endif

void	cmdscan();
SIG_RET_TYPE	getchild();
struct	sockaddr_in in_soc = { AF_INET };
int	errno;
int	flags;
int	debug;

/*
 * MAXCMDLEN used to be NCARGS, which Sun helpfully cranked up to a Meg or
 * so...
 */
#define MAXCMDLEN	8192

/* All we need for the track command arg vector is
 *  1  The command name
 * 2-3 Mandatory -g flag and host name
 * [4] Optional  -q flag
 * [5] Optional  -s flag
 * 6-7 -m DEF_ADM flag to mail errors
 *  8  Subscription list name
 *  +  NULL arg vector terminator
 *
 * If you add more arguments, enlarge the vector.
 * We'll at least check it in case you forget.
 */

#define MAXARGS 32
char	*cmdargs[MAXARGS + 1];
int	nargs;
#define addarg(x) \
	if (nargs < MAXARGS) \
		cmdargs[nargs++] = (x); \
	else \
		die("Too many arguments")

#ifdef DEBUG
#define TRACKCMDFMT	"%s/track"
#define TRACKROOTDIR	SRCDIR
#else
#define TRACKCMDFMT	"%s/bin/track"
#define TRACKROOTDIR	DEF_ROOT
#endif

char	trackcmd[sizeof(TRACKCMDFMT) + sizeof(TRACKROOTDIR)];

univptr_t
emalloc(n)
int n;
{
	char *cp = malloc((unsigned) n);
	if (cp == (char *) NULL) {
#define ERRFMT	"out of memory on request for %d bytes"
		char errbuf[sizeof(ERRFMT) + 16];

		sprintf(errbuf, ERRFMT, n);
		logmessage(errbuf);
		exit(2);
#undef ERRFMT
	}
	return cp;
}

logmessage(s)
char *s;
{
	static int logopen = 0;
	
	if (!debug && logopen == 0) {
#ifdef LOG_DAEMON
		openlog("trackd", LOG_PID, LOG_DAEMON);
#else
		openlog("trackd", LOG_PID);
#endif
		logopen++;
	}
	if (debug) {
		if (errno)
			perror("trackd");
		fprintf(stderr, "trackd: %s\n", s);
	} else
		syslog(LOG_NOTICE, errno ? "%s: %m" : "%s", s);
	errno = 0;
}

die(s)
char *s;
{
	logmessage(s);
	logmessage("trackd dying");
	closelog();
	ERROR_EXIT(1);
}

child_die(s)
char *s;
{
	logmessage(s);
	logmessage("trackd child dying");
	closelog();
	ERROR_EXIT(1);
}

eclose(fd, where)
int fd;
char	*where;
{
	int	serrno;
	if (close(fd) != 0) {
#define ERRFMT "closing file %3d (errno %3d) (from %.20s)"
		char	buffer[sizeof(ERRFMT) + 3 + 3 + 25];
		serrno = errno;
		sprintf(buffer, ERRFMT, fd, errno, where);
#undef ERRFMT
		logmessage(buffer);
		errno = serrno;
		if (errno != EINTR)
			child_die("error closing file descriptor");
/*		else
			(void) close(fd); */
	}
}

/*
 * 	protocol format:
 *	remuser\0	 not used, always run as root 
 *	locuser\0
 *	command\0
 *	data
 */
int
main(argc, argv)
int argc;
char **argv;
{
	int fid, stime, the_tty, portnum;
	struct servent *sptr;

	argc--;
	argv++;
	if (argc > 0 && !strcmp(argv[0], "-d"))
	{
		flags |= SO_DEBUG;
		argc--;
		argv++;
		debug++;
	}
	if (argc > 0 && !strcmp(argv[0], "-inetd"))
		/* the rest of the code assumes the net fd is > 2, so dup() */
		return bigloop(dup(0));
	if (debug == 0) {
		int pid = fork();
		if (pid == -1) {
		    die("fork failed");
		} else if (pid != 0) {
		    /* parent exits right away */
		    exit(0);
		}
		for (fid = 0; fid < 10; fid++) {
		    (void) close(fid);
		}
		if (open("/", 0) == -1 || dup2(0, 1) == -1 || dup2(0, 2) == -1)
			die("open(\"/\",0) or dup failed");
		the_tty = open("/dev/tty", 2);
		if (the_tty >= 0)
		{
#ifdef TIOCNOTTY
			if (ioctl(the_tty, TIOCNOTTY, (char *)0) == -1)
				die("TIOCNOTTY ioctl failed");
#endif
			(void) eclose(the_tty, "tty setup");
			setsid();
		}
	}
	errno = 0;
	if ((sptr = getservbyname(TRACKSERVICE, "tcp")) == NULL)
		die("unknown service");
	in_soc.sin_port = sptr->s_port;
	if (argc > 0)
	{
		portnum = atoi(argv[0]);
		if (portnum < 0)
			die("bad port #");
		in_soc.sin_port = htons((u_short)portnum);
		argc--, argv++;
	}

	fid = socket(AF_INET, SOCK_STREAM, 0);
	if (fid < 0)
		die("socket");
#ifdef SO_DEBUG
	{
	struct linger l;
	int on = 1;
	if ((flags & SO_DEBUG) && setsockopt(fid, SOL_SOCKET, SO_DEBUG, (char *)&on, sizeof on) < 0)
		die("setsockopt (SO_DEBUG)");
	if (setsockopt(fid, SOL_SOCKET, SO_KEEPALIVE, (char *)&on, sizeof on) < 0)
		die("setsockopt (SO_KEEPALIVE)");

	l.l_onoff = 1;
	l.l_linger = 60;
	if (setsockopt(fid, SOL_SOCKET, SO_LINGER, (char *)&l, sizeof l) < 0)
		die("setsockopt (SO_LINGER)");
	}
#endif
	if (bind(fid, (struct sockaddr *) &in_soc, sizeof (in_soc)) < 0)
		die("bind");

	signal(SIGCHLD, getchild);
	(void) listen(fid, 10);
	for (;;) {
#define MSGFMT	"Track connection from %s"
		int g, pid;

		g = accept(fid, (struct sockaddr *)0, (int *)0);
		if (g < 0) {
			if (errno != EINTR)
				logmessage("accept failed");
			else
				logmessage("Hmm - got an EINTR again");
			continue;
		}
		if ((pid = fork()) == 0) {
			/* We fork again and have the parent exit right away.
			   This winds up with the child that does all the
			   work being init's problem. */
			if ((pid = fork()) == 0) {
				signal(SIGCHLD, SIG_IGN);
				eclose(fid, "child fid");
				bigloop(g);
			}
			else if (pid == -1)
				logmessage("child fork failed");
			exit(0);
		} else if (pid == -1) {
			logmessage("fork failed");
		}
		(void) wait(0);
		eclose(g, "parent g");
	}
	/* NOTREACHED */
}

SIG_RET_TYPE
getchild()
{

	while (waitpid(-1, (int *)0, WNOHANG) > 0)
		;
}

char	env_user[] = "USER=root";
char	env_shell[] = "SHELL=/bin/sh";
char	env_home[] = "HOME=/";
char	*env_str[] =
	    {env_home,
	    env_shell,
	    "PATH=/usr/ucb:/bin:/usr/bin",
	    env_user,
	    0};

bigloop(rem)
int rem;
{
	int a_one = 1;
	int the_soc = -1, the_tty;
	int pv[2], pid, width, cc;
	fd_set readfrom, ready;
#define FDSET0 ((struct fd_set *) 0)
	short port;
	char cmdbuf[MAXCMDLEN+1];
	char locuser[16], remuser[16];
	char buf[BUFSIZ], sig, *remhname;
	char c;
	struct hostent *hp;
	struct sockaddr_in ibuf, *in_ptr = &ibuf;
	extern char *getrealpeername();

	if ((remhname = getrealpeername(rem, in_ptr)) == NULL)
	    child_die("Couldn't get name of remote host");

	if (debug) {
		sprintf(buf, MSGFMT, remhname);
		errno = 0;
		logmessage(buf);
	}

	/*
	 * !! We probably want some authentication here. Ideally, read
	 * a file to see if we're talking to this host, and if we are
	 * allowed to use this track slist
	 */

	signal(SIGINT, SIG_DFL);
	signal(SIGQUIT, SIG_DFL);
	signal(SIGTERM, SIG_DFL);

#ifdef DEBUG
	the_tty = open("/dev/tty", 2);
	if (the_tty >= 0)
	{
		if (ioctl(the_tty, TIOCNOTTY, (char *)0) == -1)
			child_die("TIOCNOTTY ioctl failed");
		eclose(the_tty, "child tty");
	}
	
#endif
	in_ptr->sin_port = ntohs((u_short)in_ptr->sin_port);
	if (in_ptr->sin_family != AF_INET)
		child_die("from address is not AF_INET");
	if (in_ptr->sin_port >= IPPORT_RESERVED ||
	    in_ptr->sin_port < IPPORT_RESERVED/2)
		child_die("connection from illegal port");
	alarm(60);
	port = 0;
	for (;;)
	{
		if (read(rem, &c, 1) != 1)
		{
			shutdown(rem, 1+1);
			child_die("read");
		}
		if (c == 0)
		{
			break;
		}
		port = port * 10 + c - '0';
	}
	alarm(0);
	if (port != 0)
	{
		int lport = IPPORT_RESERVED - 1;

		the_soc = rresvport(&lport);
		if (the_soc < 0)
		{
			child_die("can't get stderr port");
		}
		if (port >= IPPORT_RESERVED)
		{
			child_die("2nd port not reserved");
		}
		in_ptr->sin_port = htons((u_short)port);
		if (connect(the_soc, (struct sockaddr *) in_ptr,
		 sizeof (*in_ptr)) < 0)
		{
			child_die("connect");
		}
	}
	if (dup2(rem, 0) == -1 || dup2(rem, 1) == -1 || dup2(rem, 2) == -1)
		child_die("some dup failed");
	hp = gethostbyaddr((char *) &in_ptr->sin_addr, sizeof (struct in_addr),
	 in_ptr->sin_family);
	if (hp == 0)
		child_die("Host name for your address unknown");

	/*
	**	you can throw out the users, we will always run as root
	*/
	getstr(remuser, sizeof(remuser), "remuser");
	getstr(locuser, sizeof(locuser), "locuser");
	getstr(cmdbuf, sizeof(cmdbuf), "command");
	/*
	**	the librarian can always run as root
	*/

	(void) write(2, "\0", 1);
	if (port)
	{
		if (pipe(pv) < 0)
		{
			child_die("Can't make pipe.");
		}
		pid = fork();
		if (pid == -1) 
		{
			child_die("fork failed - try again.");
		}
		if (pid)
		{
			eclose(rem, "child rem");
			eclose(pv[1], "child pv[1]");
			eclose(0, "child fd 0");
			eclose(1, "child fd 1");
			eclose(2, "child fd 2");

			FD_ZERO(&readfrom);
			FD_SET(the_soc, &readfrom);
			FD_SET(pv[0], &readfrom);

#ifdef O_NONBLOCK
			fcntl(pv[0], F_SETFL, O_NONBLOCK);
#else
			(void) ioctl(pv[0], FIONBIO, (char *)&a_one);
#endif
			width = the_soc;
			if (width < pv[0]) width = pv[0];
			width++;

			do
			{
				ready = readfrom;
				if (select(width, &ready, FDSET0, FDSET0,
					    (struct timeval *)0) < 0)
				{
					break;
				}
				if (FD_ISSET(the_soc, &ready))
				{
					if (read(the_soc, &sig, 1) <= 0)
					{
						FD_CLR(the_soc, &readfrom);
					}
					else
					{
						kill(-pid, sig);
					}
				}
				if (FD_ISSET(pv[0], &ready))
				{
					errno = 0;
					cc = read(pv[0], buf, sizeof (buf));
					if (cc <= 0)
					{
						shutdown(the_soc, 1+1);
						FD_CLR(pv[0], &readfrom);
					}
					else
					{
						(void) write(the_soc, buf, cc);
					}
				}
			} while (FD_ISSET(the_soc, &readfrom) ||
			         FD_ISSET(pv[0], &readfrom));
			exit(0);
		}
		setpgrp(0, getpid());
		eclose(pv[0], "child pv[0]");
		eclose(the_soc, "child socket");
		if (dup2(pv[1], 2) == -1)
			child_die("dup failed");
	}
	sprintf(trackcmd, TRACKCMDFMT, TRACKROOTDIR);
	nargs = 0;
	addarg(trackcmd);
	addarg("-g");
	addarg(hp->h_name);

	/*
	 * Diagnostics produced by the remote track
	 * confuse the local one, so workaround the problem
	 * by mailing such to DEF_ADM.  Blech.
	 */
	addarg("-m");
	addarg(DEF_ADM);

	cmdscan(cmdbuf);
	addarg((char *) NULL);
	if (close(rem) == -1)
		child_die("couldn't close remote");
	execve(trackcmd, cmdargs, env_str);
	child_die(trackcmd);
}

getstr(ptr, cnt, err)
char *ptr;
int cnt;
char *err;
{
	char c;
	do
	{
		if (read(0, &c, 1) != 1)
			exit(1);
		*ptr++ = c;
		if (--cnt == 0) {
#define ERRMSG		" too long\n"
			char errbuf[LINELEN + sizeof(ERRMSG)];
			strncpy(errbuf, err, LINELEN);
			strcat(errbuf, ERRMSG);
			child_die(errbuf);
#undef ERRMSG
		}
	} while (c != 0);
}

/*
**	allowable flags are -s and -q
**	the first word that doesn't begin with -
**	is taken to be the subscription list
**	all other words are ignored.
**	be sure to change MAXARGS if you allow more arguments
**	command line may be modified
*/
void
cmdscan(to)
register char *to;
{
	register char *from;
	int qseen = 0;
	int sseen = 0;
	char *slist = NULL;
	char *errbuf;


	while(*to != '\0')
	{
		for (from=to; isspace(*from); from++);
		for (to=from; *to && !isspace(*to); to++);
		if (*from == '-') {
			if ((to - from) == 2) switch (from[1]) {
			case 'q': qseen = 1; continue;
			case 's': sseen = 1; continue;
			}
			/* If we get here, something's wrong */
#define ERRFMT "unknown option %s"
			errbuf = emalloc(sizeof (ERRFMT) + to - from);
			*to = '\0';
			sprintf(errbuf, ERRFMT, from);
#undef ERRFMT
			child_die(errbuf);
		}
		*to = '\0';
		slist = from;
		break;
	}
	if (qseen) addarg("-q");
	if (sseen) addarg("-s");
	if (slist != NULL) addarg(slist);
}

/*
 * Calls getpeername(s) to get the rank and serial number of the socket (but
 * not the name - of course not - why would a routine called getpeername()
 * ever return the name...), then calls gethostbyaddr() to get the real name,
 * returns NULL if it fails in any way.
 */
char *
getrealpeername(s, ia)
int s;
struct sockaddr_in *ia;
{
	struct sockaddr_in sin;
	struct hostent *hp;
	int sin_len;

	sin_len = sizeof(sin);
	if (getpeername(s, (struct sockaddr *) &sin, &sin_len) == -1)
		return NULL;
	*ia = sin;
	hp = gethostbyaddr((char *)&sin.sin_addr,
		sizeof (struct in_addr), AF_INET);
	if (hp == NULL) {
#define ERRFMT "unkown host # %20s"
		char buffer[sizeof(ERRFMT) + 20];
		char *addr = inet_ntoa(sin.sin_addr);
		sprintf(buffer, ERRFMT, addr? addr : "NULL");
#undef ERRFMT
		logmessage(buffer);
	}
	return (hp ? hp->h_name : NULL);
}
